"""
Module for downloading and listing JSON DRS manifest and DRS objects. The main classes in
this module for downloading DRS objects are DownloadManager and Manifest.

    Examples:
        This generates the Gen3Jobs class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> datafiles = Manifest.load('sample/manifest_1.json')
            downloadManager = DownloadManager("source.my_commons.org",
                              Gen3Auth(refresh_file="~.gen3/my_credentials.json"), datafiles)
            for i in datafiles:
                print(i)
            downloadManager.download(datafiles, ".")

        See docs/howto/drsDownloading.md for more details

"""


import re
import os
from dataclasses import dataclass, field
from datetime import datetime, timezone, timedelta
from enum import Enum
from json import load as json_load, loads as json_loads, JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import humanfriendly
import requests
import zipfile
from cdislogging import get_logger
from dataclasses_json import dataclass_json, LetterCase, Undefined
from dateutil import parser as date_parser
from tqdm import tqdm
from urllib.parse import urlparse

from gen3.auth import Gen3Auth, Gen3AuthError, decode_token
from gen3.auth import _handle_access_token_response
from gen3.tools.download.drs_resolvers import resolve_drs
from gen3.utils import remove_trailing_whitespace_and_slashes_in_url
from gen3.metadata import Gen3Metadata

DEFAULT_EXPIRE: timedelta = timedelta(hours=1)

# package formats we handle for unpacking
PACKAGE_EXTENSIONS = [".zip"]

logger = get_logger("__name__")


# Add undefined=Undefined.EXCLUDE here because we only cares if the input manifest has the minimal required metadata fields for data download, any extra metadata fields should be ignored and should not cause a failure
@dataclass_json(letter_case=LetterCase.SNAKE, undefined=Undefined.EXCLUDE)
@dataclass
class Manifest:
    """Data class representing a Gen3 JSON manifest typically exported from a Gen3 discovery page.

    The class is passed to the DownloadManager to download or list all of the files in the manifest.
    The Download manager will cache additional information (if available)

    Attributes:
        object_id(str): the DRS object id. This is the only attribute that needs to be defined
        file_size(Optional[int]): the filesize of the object, if contained in the manifest
        file_name(Optional[str]): the name of the file pointed to by the DRS object id
        md5sum(Optional[str]): the checksum of the object
        commons_url(Optional[str]): url of the indexd server to retrieve file/bundle from
    """

    object_id: str  # only required member
    file_size: Optional[int] = -1  # -1 indicated not set
    file_name: Optional[str] = None
    md5sum: Optional[str] = None
    commons_url: Optional[str] = None

    @staticmethod
    def load_manifest(path: Path):
        """Loads a json manifest"""
        with open(path, "rt") as fin:
            data = json_load(fin)
            return Manifest.schema().load(data, many=True)

    @staticmethod
    def create_object_list(manifest) -> List["Downloadable"]:
        """Create a list of Downloadable instances from the manifest

        Args:
            manifest(list): list of manifest objects
        Returns:
            List of Downloadable instances
        """
        results = []
        for entry in manifest:
            results.append(
                Downloadable(
                    object_id=entry.object_id,
                    hostname=remove_trailing_whitespace_and_slashes_in_url(
                        entry.commons_url
                    ),
                )
            )
        return results

    @staticmethod
    def load(filename: Path) -> Optional[List["Downloadable"]]:
        """
        Method to load a json manifest and return a list of Bownloadable object.
        This list is passed to the DownloadManager methods of download, and list

        Args:
            filename(Path): path to manifest file
        Returns:
            list of Downloadable objects if successfully opened/parsed None otherwise
        """
        try:
            manifest = Manifest.load_manifest(filename)
            return Manifest.create_object_list(manifest)
        except FileNotFoundError as ex:
            logger.critical(f"Error file not found: {ex.filename}")
        except JSONDecodeError as ex:
            logger.critical(f"format of manifest file is valid JSON: {ex.msg}")

        return None


@dataclass
class KnownDRSEndpoint:
    """
    Dataclass used internally by the DownloadManager class to cache hostnames and tokens
    for Gen3 commons possibly accessed using the Workspace Token Service (WTS). The endpoint is
    assumed to support DRS and therefore caches additional DRS information.

    Attributes:
        hostname (str): hostname of DRS server
        last_refresh_time (datetime): last time the token has been refreshed
        idp (Optional[str]): cached idp information
        access_token (Optional[str]): current bearer token initially None
        identifier (Optional[str]): DRS prefix (if used)
        use_wts (bool): if True use WTS to create tokens

    """

    hostname: str
    expire: datetime = None
    idp: Optional[str] = None
    access_token: Optional[str] = None
    identifier: Optional[str] = None
    use_wts: bool = True

    @property
    def available(self):
        """If endpoint has access token it is available"""
        return self.access_token is not None

    def expired(self) -> bool:
        """check if WTS token is older than the default expiration date

        If not using the WTS return false and use standard Gen3 Auth
        """
        if not self.use_wts:
            return False

        return datetime.now() > self.expire

    def renew_token(self, wts_server_name: str, server_access_token):
        """Gets a new token from the WTS and updates the token and refresh time

        Args:
            wts_server_name (str): hostname of WTS server
            server_access_token (str): token used to authenticate use of WTS

        """
        token = wts_get_token(
            hostname=wts_server_name,
            idp=self.idp,
            access_token=server_access_token,
        )
        token_info = decode_token(token)
        # TODO: this would break if user is trying to download object from different commons
        # keep BRH token and wts sparate
        self.access_token = token
        self.expire = datetime.fromtimestamp(token_info["exp"])


class DRSObjectType(str, Enum):
    """Enum defining the 3 possible DRS object types."""

    unknown = "unknown"
    object = "object"
    bundle = "bundle"


@dataclass
class Downloadable:
    """
    Class handling the information for a DRS object. The information is populated from the manifest or
    by retrieving the information from a DRS server.

    Attributes:
        object_id (str): DRS object id (REQUIRED)
        object_type (DRSObjectType): type of DRS object
        hostname (str): hostname of DRS object
        file_size (int): size in bytes
        file_name (str): name of file
        updated_time (datetime): timestamp of last update to file
        created_time (datetime): timestamp when file is created
        access_methods (List[Dict[str, Any]]): list of access methods (e.g. s3) for DRS object
        children (List[Downloadable]): list of child objects (in the case of DRS bundles)
        _manager (DownloadManager): manager for this Downloadable
    """

    object_id: str
    object_type: Optional[DRSObjectType] = DRSObjectType.unknown
    hostname: Optional[str] = None
    file_size: Optional[int] = -1
    file_name: Optional[str] = None
    updated_time: Optional[datetime] = None
    created_time: Optional[datetime] = None
    access_methods: List[Dict[str, Any]] = field(default_factory=list)
    children: List["Downloadable"] = field(default_factory=list)
    _manager = None

    def __str__(self):
        return (
            f'{self.file_name if self.file_name is not None else "not available" : >45}; '
            f"{humanfriendly.format_size(self.file_size) :>12}; "
            f'{self.hostname if self.hostname is not None else "not resolved"}; '
            f'{self.created_time.strftime("%m/%d/%Y, %H:%M:%S") if self.created_time is not None else "not available"}'
        )

    def __repr__(self):
        return (
            f'(Downloadable: {self.file_name if self.file_name is not None else "not available"}; '
            f"{humanfriendly.format_size(self.file_size)}; "
            f'{self.hostname if self.hostname is not None else "not resolved"}; '
            f'{self.created_time.strftime("%m/%d/%Y, %H:%M:%S") if self.created_time is not None else "not available"})'
        )

    def download(self):
        """calls the manager to download this object. Allows Downloadables to be self downloading"""
        self._manager.download([self])

    def pprint(self, indent: str = ""):
        """
        Pretty prints the object information. This is used for listing an object.
        In the case of a DRS bundle the child objects are listed similar to the linux tree command
        """

        from os import linesep

        res = self.__str__() + linesep
        child_indent = f"{indent}    "

        pos = -1
        for x in self.children:
            pos += 1
            if pos == len(self.children) - 1:
                res += f"{child_indent}└── {x.pprint(child_indent)}"
            else:
                res += f"{child_indent}├── {x.pprint(child_indent)}"
        return res


@dataclass
class DownloadStatus:
    """Stores the download status of objectIDs.

    The DataManager will return a list of DownloadStatus as a result of calling the download method

    Status is "pending" until it is downloaded or an error occurs.
    Attributes:
        filename (str): the name of the file to download
        status (str): status of file download initially "pending"
        start_time (Optional[datetime]): start time of download as datetime initially None
        end_time (Optional[datetime]): end time of download as datetime initially None
    """

    filename: str
    status: str = "pending"
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    def __str__(self):
        return (
            f'filename: {self.filename if self.filename is not None else "not available"}; '
            f"status: {self.status}; "
            f'start_time: {self.start_time.strftime("%m/%d/%Y, %H:%M:%S") if self.start_time is not None else "n/a"}; '
            f'end_time: {self.end_time.strftime("%m/%d/%Y, %H:%M:%S") if self.start_time is not None else "n/a"}'
        )

    def __repr__(self):
        return self.__str__()


def wts_external_oidc(hostname: str) -> Dict[str, Any]:
    """
    Get the external_oidc from a connected WTS. Will report if WTS service
    is missing. Note that in some cases this can be considered a warning not a error.

    Args:
        hostname (str): hostname to access the WTS endpoint

    Returns:
        dict containing the oidc information
    """
    oidc = {}
    if not hostname:
        return oidc
    try:
        response = requests.get(f"https://{hostname}/wts/external_oidc/")
        response.raise_for_status()
        data = response.json()
        if "providers" not in data:
            logger.warning(
                'cannot find "providers". Likely no WTS service running for this commons'
            )
            return oidc
        for item in data["providers"]:
            oidc[urlparse(item["base_url"]).netloc] = item

    except requests.exceptions.HTTPError as exc:
        logger.critical(
            f'HTTP Error ({exc.response.status_code}): {json_loads(exc.response.text).get("message", "")}'
        )
    except JSONDecodeError as ex:
        logger.warning(
            f"Unable to process WTS response. Likely no WTS service running on this commons. "
            f"Certain commands might fail."
        )

    return oidc


def wts_get_token(hostname: str, idp: str, access_token: str):
    """
    Gets a auth token from a Gen3 WTS server for the supplied idp
    Args:
        hostname (str): Gen3 common's WTS service
        idp: identity provider to use
        access_token: Gen3 Auth to use to with WTS

    Returns:
        Token for idp if successful, None if failure
    """
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Connection": "keep-alive",
        "Authorization": "bearer " + access_token,
    }

    try:
        url = f"https://{hostname}/wts/token/?idp={idp}"
        try:
            response = requests.get(url=url, headers=headers)
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            logger.critical(
                f"HTTP Error ({exc.response.status_code}): getting WTS token: {exc.response.text}"
            )
            logger.critical(
                "Please make sure the target commons is connected on your profile page and that connection has not expired."
            )
            return None

        return _handle_access_token_response(response, "token")

    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return None


def get_drs_object_info(hostname: str, object_id: str) -> Optional[dict]:
    """
    Retrieves information for a DRS object residing on the hostname
    Args:
        hostname (str): hostname of DRS object
        object_id (str): DRS object id

    Returns:
        GAG4H DRS object information if sucessful otherwise None
    """
    try:
        response = requests.get(f"https://{hostname}/ga4gh/drs/v1/objects/{object_id}")
        response.raise_for_status()
        data = response.json()
        return data

    except requests.HTTPError as exc:
        if exc.response.status_code == 404:
            logger.critical(
                f"HTTP Error ({exc.response.status_code}): {object_id} not found at {hostname}"
            )
        else:
            logger.critical(
                f"HTTP Error ({exc.response.status_code}): accessing object: {object_id}"
            )
        return None
    except ConnectionError as exc:
        logger.critical(f"Connection Error {exc} when accessing object: {object_id}")
        return None


def extract_filename_from_object_info(object_info: dict) -> Optional[str]:
    """Extracts the filename from the object_info.

    if filename is in object_info use that, otherwise try to extract it from the one of the access methods.
    Returns filename if found, else return None

    Args
        object_info (dict): DRS object dictionary
    """
    if "name" in object_info and object_info["name"]:
        return object_info["name"]

    for access_method in object_info["access_methods"]:
        url = access_method["access_url"]["url"]
        parts = url.split("/")
        if parts:
            return parts[-1]

    return None


def get_access_methods(object_info: dict) -> List[str]:
    """
    Returns the DRS GA4GH access methods from the object_info.

    Args:
        object_info (dict): dict of GA4GH DRS Object information

    Returns:
        List of access methods
    """
    if object_info is None:
        logger.critical("no access methods defined for this file")
        return []
    return object_info["access_methods"]


def get_drs_object_type(object_info: dict) -> DRSObjectType:
    """From the object info determine the type of object.

    Args:
        object_info (dict): DRS object dictionary

    Returns:
        type of object: either bundle or object
    """
    if "form" in object_info:
        if object_info["form"] is None:
            return DRSObjectType.object
        return DRSObjectType(object_info["form"])

    if "contents" in object_info and len(object_info["contents"]) > 0:
        return DRSObjectType.bundle
    else:
        return DRSObjectType.object


def get_drs_object_timestamp(s: Optional[str]) -> Optional[datetime]:
    """returns the timestamp in datetime if not none otherwise returns None

    Args:
        s (Optional[str]): string to parse
    Returns:
        datetime if not None
    """
    return date_parser.parse(s) if s is not None else None


def add_drs_object_info(info: Downloadable) -> bool:
    """
    Given a downloader object fill in the required fields
    from the resolved hostname.
    In the case of a bundle, try to resolve all object_ids contained in the bundle including other objects and bundles.

    Args:
        info (Downloadable): Downloadable to add information to
    Returns:
         True if object is valid and resolvable.
    """
    if info.hostname is None:
        return False

    object_info = get_drs_object_info(info.hostname, info.object_id)
    if (object_info) is None:
        return False

    # Get common information we want
    info.file_name = extract_filename_from_object_info(object_info)
    info.file_size = object_info.get("size", -1)
    info.updated_time = get_drs_object_timestamp(object_info.get("updated_time", None))
    info.created_time = get_drs_object_timestamp(object_info.get("created_time", None))

    info.object_type = get_drs_object_type(object_info)
    if info.object_type == DRSObjectType.object:
        info.access_methods = get_access_methods(object_info)
        return True
    else:  # a bundle,get everything else
        for item in object_info["contents"]:
            child_id = item.get("id", None)
            if child_id is None:
                continue
            child_object = Downloadable(hostname=info.hostname, object_id=child_id)
            add_drs_object_info(child_object)
            info.children.append(child_object)

    return True


class InvisibleProgress:
    """
    Invisible progress bar which stubs a tqdm progress bar
    """

    def update(self, value):  # pragma: no cover
        pass


def download_file_from_url(
    url: str, filename: Path, show_progress: bool = True
) -> bool:
    """
    Downloads a file using the URL. The URL is a pre-signed url created by the download manager
    from the access method of the DRS object.

    Args:
        url (str): URL to download from
        filename (str): name of the file to write data to
        show_progress (bool): show a progress bar (default)

    Returns:
        True if object has been downloaded
    """
    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()

    except requests.exceptions.Timeout:
        logger.critical(f"Was unable to get the download url: {url}. Timeout Error.")
        return False
    except requests.exceptions.HTTPError as exc:
        logger.critical(
            f"HTTP Error ({exc.response.status_code}): downloading file from {url}"
        )
        return False

    total_size_in_bytes = int(response.headers.get("content-length", 0))

    if total_size_in_bytes == 0:
        logger.critical(f"content-length is 0 and it should not be")
        return False

    total_downloaded = 0
    block_size = 8092  # 8K blocks might want to tune this.

    progress_bar = (
        tqdm(
            desc=f"{str(filename) : <45}",
            total=total_size_in_bytes,
            unit="iB",
            unit_scale=True,
            bar_format="{l_bar:45}{bar:35}{r_bar}{bar:-10b}",
        )
        if show_progress
        else InvisibleProgress()
    )

    # if the file name contains '/', create subdirectories and download there
    ensure_dirpath_exists(Path(os.path.dirname(filename)))

    try:
        with open(filename, "wb") as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                total_downloaded += len(data)
                file.write(data)
    except IOError as ex:
        logger.critical(f"IOError opening {filename} for writing: {ex}")
        return False

    if total_downloaded != total_size_in_bytes:
        logger.critical(
            f"Error in downloading {filename}: expected {total_size_in_bytes} bytes, downloaded {total_downloaded} bytes"
        )
        return False
    return True


def unpackage_object(filepath: str):
    # allowed formats are set in PACKAGE_EXTENSIONS
    with zipfile.ZipFile(filepath, "r") as package:
        package.extractall(os.path.dirname(filepath))


def parse_drs_identifier(drs_candidate: str) -> Tuple[str, str, str]:
    """
    Parses a DRS identifier to extract a hostname in the case of hostname based DRS
    otherwise it look for a DRS compact identifier.

    If neither one is recognized return an empty string and a type of 'unknown'
    Note: The regex expressions used to extract hostname or identifier has a potential for
    a false positive.

    Args:
        drs_candidate (str): a drs object identifier

    Returns:
        Tuple (str): tuple of hostname/drs prefix, guid, string: one of "hostname", "compact", "unknown"
    """
    # determine if hostname or compact identifier or unknown
    drs_regex = r"drs://([A-Za-z0-9\.\-\~]+)/([A-Za-z0-9\.\-\_\~\/]+)"
    # either a drs prefix:

    matches = re.findall(drs_regex, drs_candidate, re.UNICODE)

    if len(matches) == 1:  # this could be a hostname DRS id
        hostname_regex = (
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*"
            r"([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
        hostname_matches = re.findall(hostname_regex, matches[0][0], re.UNICODE)
        if len(hostname_matches) == 1:
            return matches[0][0], matches[0][1], "hostname"
    # possible compact rep
    compact_regex = r"([A-Za-z0-9\.\-\~]+)/([A-Za-z0-9\.\-\_\~\/]+)"
    matches = re.findall(compact_regex, drs_candidate, re.UNICODE)
    if len(matches) == 1 and len(matches[0]) == 2:
        return matches[0][0], matches[0][1], "compact"

    # can't figure out a this identifier
    return "", "", "unknown"


def resolve_drs_hostname_from_id(
    object_id: str, resolved_drs_prefix_cache: dict, mds_url: str
) -> Optional[Tuple[str, str, str]]:
    """Resolves and returns a DRS identifier
    The resolved_drs_prefix_cache is updated if needed and is a potential side effect of this
    call
    Args:
        object_id (str): DRS object id to resolve
        resolved_drs_prefix_cache (dict) : cache of resolved DRS prefixes
        mds_url (str): the URL for the Gen3 Aggregate MDS to use to help resolved DRS hostname

    Returns:
         the hostname of the DRS server if resolved, otherwise it returns None
    """
    hostname = None

    prefix, identifier, identifier_type = parse_drs_identifier(object_id)
    if identifier_type == "hostname":
        return prefix, identifier, identifier_type
    if identifier_type == "compact":
        if prefix not in resolved_drs_prefix_cache:
            hostname = resolve_drs(prefix, object_id, metadata_service_url=mds_url)
            if hostname is not None:
                resolved_drs_prefix_cache[prefix] = hostname
        else:
            hostname = resolved_drs_prefix_cache[prefix]

    return hostname, identifier, identifier_type


def resolve_objects_drs_hostname(
    object_ids: List[Downloadable],
    resolved_drs_prefix_cache: dict,
    mds_url: str,
    commons_url: str = None,
) -> None:
    """Given a list of object_ids go through list and resolve + cache any unknown hosts

    Args:
        object_ids (List[Downloadable]): list of object to resolve
        resolved_drs_prefix_cache (dict): cache of resolved DRS prefixes
        mds_url (str): Gen3 metadata service to resolve DRS prefixes
        hostname (str): Hostname to main Gen3 environment
    """
    for entry in object_ids:
        if commons_url is not None:
            entry.hostname = commons_url
        if entry.hostname is None:
            # if resolution fails the entry hostname will still be None
            entry.hostname, nid, drs_type = resolve_drs_hostname_from_id(
                entry.object_id, resolved_drs_prefix_cache, mds_url
            )
            if (
                drs_type == "hostname"
            ):  # drs_type is a hostname so object id will be the GUID
                entry.object_id = nid


def ensure_dirpath_exists(path: Path) -> Path:
    """Utility to create a directory if missing.
    Returns the path so that the call can be inlined in a another call

    Args:
        path (Path): path to create
    Returns
        path of created directory
    """
    assert path
    out_path: Path = path

    if not out_path.exists():
        out_path.mkdir(parents=True, exist_ok=True)

    return out_path


def get_download_url_using_drs(
    drs_hostname: str, object_id: str, access_method: str, access_token: str
) -> Optional[str]:
    """
    Returns the presigned URL for a DRS object, from a DRS hostname, via the access method
    Args:
        drs_hostname (str): hostname of DRS server
        object_id (str): DRS object id
        access_method (str): access method to use
        access_token (str): access token to DRS server

    Returns:
        presigned url to object
    """
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "bearer " + access_token,
    }
    try:
        response = requests.get(
            url=f"https://{drs_hostname}/ga4gh/drs/v1/objects/{object_id}/access/{access_method}",
            headers=headers,
        )
        response.raise_for_status()
        data = response.json()
        return data.get("url", None)

    except requests.exceptions.Timeout:
        logger.critical(f"Was unable to download: {object_id}. Timeout Error.")
    except requests.exceptions.HTTPError as exc:
        logger.critical(
            f"HTTP Error ({exc.response.status_code}) when requesting download url from {access_method}"
        )
    return None


def get_user_auth(commons_url: str, access_token: str) -> Optional[List[str]]:
    """
    Retrieves a user's authz for the commons based on the access token.
    Any error will be logged and None is returned

    Args:
        commons_url (str): hostname of Gen3 indexd
        access_token (str): user's auth token

    Returns:
        The authz object from the user endpoint
    """
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "bearer " + access_token,
    }

    try:
        response = requests.get(url=f"https://{commons_url}/user/user", headers=headers)
        response.raise_for_status()
        data = response.json()
        authz = data["authz"]
        return authz
    except requests.exceptions.HTTPError as exc:
        logger.critical(f"HTTP Error ({exc.response.status_code}): getting user access")

    return None


def list_auth(hostname: str, authz: dict):
    """
    Prints the authz for a DRS hostname
    Args:
        hostname (str): hostname to list access
        authz (str): dictionary of authz stringts
    """
    print(
        "───────────────────────────────────────────────────────────────────────────────────────────────────────"
    )
    print(f"Access for {hostname}:")
    if authz is not None and len(authz) > 0:
        for access, methods in authz.items():
            print(
                f"      {access : <55}: {' '.join(dict.fromkeys(x['method'] for x in methods).keys()):>40}"
            )
    else:
        print("      No access")


def get_hostname_from_endpoint(endpoint: str):
    """
    Get hostname from an Gen3Auth endpoint value
    Args:
        endpoint (str): endpoint value form Gen3Auth
    """
    if not endpoint:
        return None
    urlparts = urlparse(endpoint)
    return urlparts.hostname


class DownloadManager:
    """
    Class to assist in downloading a list of Downloadable object which at a minimum is a json manifest
    of DRS object ids. The methods of interest are download and user_access.
    """

    def __init__(
        self,
        hostname: str,
        auth: Gen3Auth,
        download_list: List[Downloadable],
        show_progress: bool = False,
        commons_url: str = None,
    ):
        """
        Initialize the DownloadManager so that is ready to start downloading.
        Note the downloadable objects are required so that all tokens are available
        to support the download.

        Args:
            hostname (str): Gen3 commons home commons
            auth (Gen3Auth) : Gen3 authentication
            download_list (List[Downloadable]): list of objects to download
        """

        self.hostname = (
            hostname if hostname else get_hostname_from_endpoint(auth.endpoint)
        )
        self.commons_url = commons_url
        self.access_token = auth.get_access_token()
        self.metadata = Gen3Metadata(auth)
        self.wts_endpoints = wts_external_oidc(hostname)
        self.resolved_compact_drs = {}
        # add COMMONS host as a DRSEndpoint as it does not use the WTS
        self.known_hosts = {
            self.hostname: KnownDRSEndpoint(
                hostname=self.hostname,
                access_token=self.access_token,
                use_wts=False,
                expire=datetime.fromtimestamp(decode_token(self.access_token)["exp"]),
            )
        }
        self.download_list = download_list
        self.resolve_objects(self.download_list, show_progress)

    def resolve_objects(self, object_list: List[Downloadable], show_progress: bool):
        """
        Given an Downloadable object list, resolve the DRS hostnames and update each Downloadable

        Args:
            object_list (List[Downloadable]): list of Downloadable objects to resolve
        """
        resolve_objects_drs_hostname(
            object_list,
            self.resolved_compact_drs,
            mds_url=f"http://{self.hostname}/mds/aggregate/info"
            if self.hostname
            else None,
            commons_url=self.commons_url,
        )
        progress_bar = (
            tqdm(desc=f"Resolving objects", total=len(object_list))
            if show_progress
            else InvisibleProgress()
        )
        for entry in object_list:
            add_drs_object_info(entry)
            # sugar to allow download objects to self download
            entry._manager = self
            progress_bar.update(1)

    def cache_hosts_wts_tokens(self, object_list):
        """
        Using the list of DRS host obtain a WTS token for all DRS hosts in the list. It's is possible
        """
        # create two sets: one of the know WTS host and the other of the host in the manifest
        wts_endpoint_set = set(self.wts_endpoints.keys())
        object_id_hostnames = {
            x.hostname for x in object_list if x.hostname is not None
        }

        drs_in_wts = wts_endpoint_set.intersection(
            object_id_hostnames
        )  # all DRS host in WTS
        drs_not_in_wts = object_id_hostnames.difference(
            wts_endpoint_set
        )  # all DRS host not in WTS
        for drs_hostname in drs_in_wts:
            endpoint = KnownDRSEndpoint(
                hostname=drs_hostname,
                idp=self.wts_endpoints[drs_hostname]["idp"],
            )
            endpoint.renew_token(self.hostname, self.access_token)
            self.known_hosts[drs_hostname] = endpoint
        for drs_hostname in drs_not_in_wts:
            # if we already know the host then we don't need to reset the host
            if drs_hostname in self.known_hosts:
                continue
            # mark hostname as unavailable
            self.known_hosts[drs_hostname] = KnownDRSEndpoint(
                hostname=drs_hostname,
            )
            logger.critical(
                f"Could not retrieve a token for {drs_hostname}: it is not available as a WTS endpoint."
            )

    def get_fresh_token(self, drs_hostname: str) -> Optional[str]:
        """Will return and/or refresh and return a WTS token if hostname is known otherwise returns None.
        Args:
            drs_hostname (str): hostname to get token for

        Returns:
            access token if successful otherwise None
        """
        if drs_hostname not in self.known_hosts:
            logger.critical(f"Could not find {drs_hostname} in cache.")
            return None
        if self.known_hosts[drs_hostname].available:
            if not self.known_hosts[drs_hostname].expired():
                return self.known_hosts[drs_hostname].access_token
            else:
                # update the token
                self.known_hosts[drs_hostname].renew_token(
                    self.hostname, self.access_token
                )
                return self.known_hosts[drs_hostname].access_token

        return None

    def download(
        self,
        object_list: List[Downloadable],
        save_directory: str = ".",
        show_progress: bool = False,
        unpack_packages: bool = True,
        delete_unpacked_packages: bool = False,
    ) -> Dict[str, Any]:
        """
        Downloads objects to the directory or current working directory.
        The input is an list of Downloadable object created by loading a manifest
        using the Manifest class or a call to Manifest.load(...

        The download manager will download each file in the manifest, in the
        case of errors they are logged and it continues.

        The return value is a list of DownloadStatus object, detailing the results
        of the download.

        Args:
            object_list (List[Downloadable]):
            save_directory (str): directory to save to (will be created)
            show_progress (bool): show a download progress bar
            unpack_packages (bool): set to False to disable the unpacking of downloaded packages
            delete_unpacked_packages (bool): set to True to delete package files after unpacking them

        Returns:
            List of DownloadStatus objects for each object id in object_list
        """

        self.cache_hosts_wts_tokens(object_list)
        output_dir = Path(save_directory)

        completed = {
            entry.object_id: DownloadStatus(filename=entry.file_name)
            for entry in object_list
        }

        for entry in object_list:
            # handle bundles first
            if entry.object_type is DRSObjectType.bundle:
                # append the filename to the directory path and
                child_dir = Path(save_directory, entry.file_name)
                # call download with the children object list
                child_status = self.download(
                    entry.children,
                    child_dir,
                    show_progress,
                    unpack_packages,
                    delete_unpacked_packages,
                )
                # when complete, append the return status
                completed[entry.object_id] = child_status
                continue

            if entry.hostname is None:
                logger.critical(
                    f"{entry.hostname} was not resolved, skipping {entry.object_id}."
                    f"Skipping {entry.file_name}"
                )
                completed[entry.object_id].status = "error (resolving DRS host)"
                continue

            # check to see if we have tokens
            if entry.hostname not in self.known_hosts:
                logger.critical(
                    f"{entry.hostname} is not present in this commons remote user access."
                    f"Skipping {entry.file_name}"
                )
                completed[entry.object_id].status = "error (resolving DRS host)"
                continue
            if self.known_hosts[entry.hostname].available is False:
                logger.critical(
                    f"Was unable to get user authorization from {entry.hostname}. Skipping {entry.file_name}"
                )
                completed[entry.object_id].status = "error (no auth)"
                continue

            drs_hostname = entry.hostname
            access_token = self.get_fresh_token(drs_hostname)

            if access_token is None:
                logger.critical(
                    f"No access token defined for {entry.object_id}. Skipping"
                )
                completed[entry.object_id].status = "error (no access token)"
                continue
            # TODO refine the selection of access_method
            if len(entry.access_methods) == 0:
                logger.critical(
                    f"No access methods defined for {entry.object_id}. Skipping"
                )
                completed[entry.object_id].status = "error (no access methods)"
                continue
            access_method = entry.access_methods[0]["access_id"]

            download_url = get_download_url_using_drs(
                drs_hostname,
                entry.object_id,
                access_method,
                access_token,
            )

            if download_url is None:
                completed[entry.object_id].status = "error"
                continue

            completed[entry.object_id].start_time = datetime.now(timezone.utc)
            filepath = output_dir.joinpath(entry.file_name)
            res = download_file_from_url(
                url=download_url, filename=filepath, show_progress=show_progress
            )

            # check if the file is a package; if so, unpack it in place
            ext = os.path.splitext(entry.file_name)[-1]
            if unpack_packages and ext in PACKAGE_EXTENSIONS:
                try:
                    mds_entry = self.metadata.get(entry.object_id)
                except Exception:
                    mds_entry = {}  # no MDS or object not in MDS
                    logger.debug(
                        f"{entry.file_name} is not a package and will not be expanded"
                    )

                # if the metadata type is "package", then unpack
                if mds_entry.get("type") == "package":
                    try:
                        unpackage_object(filepath)
                    except Exception as e:
                        logger.critical(
                            f"{entry.file_name} had an issue while being unpackaged: {e}"
                        )
                        res = False

                    if delete_unpacked_packages:
                        filepath.unlink()
            if res:
                completed[entry.object_id].status = "downloaded"
                logger.debug(
                    f"object {entry.object_id} has been successfully downloaded."
                )
            else:
                completed[entry.object_id].status = "error"
                logger.debug(f"object {entry.object_id} has failed to be downloaded.")
            completed[entry.object_id].end_time = datetime.now(timezone.utc)

        return completed

    def user_access(self):
        """
        List the user's access permissions on each host needed to download
        DRS objects in the manifest. A useful way to determine if access permissions
        are one reason a download failed.

        Returns:
            list of authz for each DRS host
        """
        results = {}
        self.cache_hosts_wts_tokens(self.download_list)
        for hostname in self.known_hosts.keys():
            if self.known_hosts[hostname].available is False:
                logger.critical(
                    f"Was unable to get user authorization from {hostname}."
                )
                continue
            access_token = self.known_hosts[hostname].access_token
            authz = get_user_auth(hostname, access_token)
            results[hostname] = authz

        return results


def _download(
    hostname,
    auth,
    infile,
    output_dir=".",
    show_progress=False,
    unpack_packages=True,
    delete_unpacked_packages=False,
) -> Optional[Dict[str, Any]]:
    """
    A convenience function used to download a json manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth (str): Gen3 Auth instance
        infile (str): manifest file
        output_dir: directory to save downloaded files to
        show_progress: show progress bar
        unpack_packages (bool): set to False to disable the unpacking of downloaded packages
        delete_unpacked_packages (bool): set to True to delete package files after unpacking them

    Returns:
        List of DownloadStatus objects for each object id in object_list
    """
    object_list = Manifest.load(Path(infile))
    if object_list is None:
        logger.critical(f"Error loading {infile}")
        return None
    try:
        auth.get_access_token()
    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return

    downloader = DownloadManager(
        hostname=hostname,
        auth=auth,
        download_list=object_list,
        show_progress=show_progress,
    )

    out_dir_path = ensure_dirpath_exists(Path(output_dir))
    return downloader.download(
        object_list,
        str(out_dir_path),
        show_progress=show_progress,
        unpack_packages=unpack_packages,
        delete_unpacked_packages=delete_unpacked_packages,
    )


def _download_obj(
    hostname,
    auth,
    object_ids,
    output_dir=".",
    show_progress=False,
    unpack_packages=True,
    delete_unpacked_packages=False,
    commons_url=None,
) -> Optional[Dict[str, Any]]:
    """
    A convenience function used to download a single DRS object.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        object_ids (List[str]): DRS object id
        output_dir: directory to save downloaded files to
        show_progress: show progress bar
        unpack_packages (bool): set to False to disable the unpacking of downloaded packages
        delete_unpacked_packages (bool): set to True to delete package files after unpacking them

    Returns:
        List of DownloadStatus objects for the DRS object
    """
    try:
        auth.get_access_token()
    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return None

    object_list = [Downloadable(object_id=object_id) for object_id in object_ids]
    downloader = DownloadManager(
        hostname=hostname,
        auth=auth,
        download_list=object_list,
        show_progress=show_progress,
        commons_url=commons_url,
    )

    out_dir_path = ensure_dirpath_exists(Path(output_dir))
    return downloader.download(
        object_list,
        str(out_dir_path),
        show_progress=show_progress,
        unpack_packages=unpack_packages,
        delete_unpacked_packages=delete_unpacked_packages,
    )


def _listfiles(hostname, auth, infile: str) -> bool:
    """
    A wrapper function used by the cli to list files in a manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        infile (str): manifest file

    Returns:
        True if successfully listed
    """
    object_list = Manifest.load(Path(infile))
    if object_list is None:
        return False

    try:
        auth.get_access_token()
    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return False
    except requests.exceptions.RequestException as ex:
        logger.critical(
            f"Unable to authenticate your credentials with {hostname}: {str(ex)}"
        )
        return False

    DownloadManager(
        hostname=hostname, auth=auth, download_list=object_list, show_progress=True
    )

    for x in object_list:
        print(x.pprint())

    return True


def _list_object(hostname, auth, object_id: str) -> bool:
    """
    Lists a DRS object.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        object_id (str): DRS object

    Returns:
        True if successfully listed
    """
    try:
        auth.get_access_token()
    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return False
    except requests.exceptions.RequestException as ex:
        logger.critical(
            f"Unable to authenticate your credentials with {hostname}: {ex}"
        )
        return False

    object_list = [Downloadable(object_id=object_id)]
    DownloadManager(
        hostname=hostname, auth=auth, download_list=object_list, show_progress=False
    )

    for x in object_list:
        print(x.pprint())

    return True


def _list_access(hostname, auth, infile: str) -> bool:
    """
    A convenience function to list a users access for all DRS hostname in a manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        infile (str): manifest file

    Returns:
        True if successfully listed
    """
    object_list = Manifest.load(Path(infile))
    if object_list is None:
        return False

    try:
        auth.get_access_token()
    except Gen3AuthError:
        logger.critical(f"Unable to authenticate your credentials with {hostname}")
        return False
    except requests.exceptions.RequestException as ex:
        logger.critical(
            f"Unable to authenticate your credentials with {hostname}: {ex}"
        )
        return False

    download = DownloadManager(
        hostname=hostname, auth=auth, download_list=object_list, show_progress=False
    )
    access = download.user_access()
    for h, access in access.items():
        list_auth(h, access)

    return True


# These functions are exposed to the SDK's cli under the drs-pull subcommand
def list_files_in_drs_manifest(hostname, auth, infile: str) -> bool:
    """
    A wrapper function used by the cli to list files in a manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        infile (str): manifest file

    Returns:
        True if successfully listed
    """
    return _listfiles(hostname, auth, infile)


def list_drs_object(hostname, auth, object_id: str) -> bool:
    """
    A convenience function used to list a DRS object.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        object_id (str): DRS object

    Returns:
        True if successfully listed
    """
    return _list_object(hostname, auth, object_id)  # pragma: no cover


def download_files_in_drs_manifest(
    hostname,
    auth,
    infile,
    output_dir,
    show_progress=True,
    unpack_packages=True,
    delete_unpacked_packages=False,
) -> None:
    """
    A convenience function used to download a json manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth (str): Gen3 Auth instance
        infile (str): manifest file
        output_dir: directory to save downloaded files to
        unpack_packages (bool): set to False to disable the unpacking of downloaded packages
        delete_unpacked_packages (bool): set to True to delete package files after unpacking them

    Returns:
    """
    _download(
        hostname,
        auth,
        infile,
        output_dir,
        show_progress,
        unpack_packages,
        delete_unpacked_packages,
    )


def download_drs_objects(
    hostname,
    auth,
    object_ids,
    output_dir,
    show_progress=True,
    unpack_packages=True,
    delete_unpacked_packages=False,
    commons_url=None,
) -> None:
    """
    A convenience function used to download a single DRS object.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        object_ids (List[str]): DRS object ids
        output_dir: directory to save downloaded files to
        unpack_packages (bool): set to False to disable the unpacking of downloaded packages
        delete_unpacked_packages (bool): set to True to delete package files after unpacking them

    Returns:
        List of DownloadStatus objects for the DRS object
    """
    return _download_obj(
        hostname,
        auth,
        object_ids,
        output_dir,
        show_progress,
        unpack_packages,
        delete_unpacked_packages,
        commons_url,
    )


def list_access_in_drs_manifest(hostname, auth, infile) -> bool:
    """
    A convenience function to list a users access for all DRS hostname in a manifest.
    Args:
        hostname (str): hostname of Gen3 commons to use for access and WTS
        auth: Gen3 Auth instance
        infile (str): manifest file

    Returns:
        True if successfully listed
    """
    return _list_access(hostname, auth, infile)



import asyncio
import csv
import datetime
import json
import os
import tempfile
from urllib.parse import urlparse

import requests.exceptions
from cdislogging import get_logger

from gen3.index import Gen3Index
from gen3.metadata import Gen3Metadata
from gen3.tools import metadata
from gen3.utils import deep_dict_update

MAX_GUIDS_PER_REQUEST = 2000
MAX_CONCURRENT_REQUESTS = 5
BASE_CSV_PARSER_SETTINGS = {
    "delimiter": ",",
    "quotechar": "",
    "quoting": csv.QUOTE_NONE,
    "escapechar": "\\",
}

GUID_TYPE = "subject"
CROSSWALK_NAMESPACE = "crosswalk"

logging = get_logger("__name__", log_level="debug")


async def read_crosswalk_metadata(
    auth, output_filename="crosswalk_metadata.csv", limit=500
):
    """
    fetch crosswalk metadata from a commons and output to crosswalk-metadata.tsv

    Data in the commons is in this format:

        "GUID": {
          "crosswalk": {
            "subject": {
              "{{commons_url}}": {
                "{{field_name}}": {
                  "value": "",
                  "type": "",
                  "description": ""
                }
                // ... more field entries here
              },
              // ... more commons entries here
              "mapping_methodologies": [
                ""
              ]
            }
          }
        }

    Output format of crosswalk data in csv:

    commons_url|identifier_type|identifier_name,commons_url|identifier_type|identifier_name
    A01-00888, phs002363.v1_RC-1358
    ...

    Output format of additional info in csv (including descriptions):
        {{commons_url}}, {{identifier_name}}, {{description}}

    See docs/howto/crosswalk.md for more detailed information.
    """
    mds = Gen3Metadata(auth_provider=auth)

    count = 0
    with tempfile.TemporaryDirectory() as metadata_cache_dir:
        crosswalk_columns = set()

        # dictionary of { "commons url | identifier name": description }
        crosswalk_info = {}

        for offset in range(0, limit, MAX_GUIDS_PER_REQUEST):
            partial_metadata = mds.query(
                f"_guid_type={GUID_TYPE}",
                return_full_metadata=True,
                limit=min(limit, MAX_GUIDS_PER_REQUEST),
                offset=offset,
            )

            if not len(partial_metadata):
                break

            # dump crosswalk metadata into temporary files so we don't have to
            # hold everything in memory for output. Do keep track of the
            # columns and field descriptions for output (that should be small)
            for guid, guid_metadata in partial_metadata.items():
                with open(
                    f"{metadata_cache_dir}/{guid.replace('/', '_')}", "w+"
                ) as cached_guid_file:
                    guid_crosswalk_metadata = guid_metadata[CROSSWALK_NAMESPACE]

                    for commons_url, commons_crosswalk in guid_crosswalk_metadata.get(
                        GUID_TYPE
                    ).items():
                        # guid_crosswalk_metadata.get(GUID_TYPE) is something like this:
                        #
                        # {
                        #   "{{commons_url}}": {
                        #     "{{field_name}}": {
                        #       "value": "",
                        #       "type": "",
                        #       "description": ""
                        #     }
                        #     // ... more field entries here
                        #   },
                        #   // ... more commons entries here
                        #   "mapping_methodologies": [
                        #     ""
                        #   ]
                        # }

                        # don't interpret mapping info as a column in the crosswalk file
                        if commons_url == "mapping_methodologies":
                            continue

                        for (
                            identifier_name,
                            indentifer_info,
                        ) in commons_crosswalk.items():
                            # identifier_name, indentifer_info is something like:
                            #
                            #     "{{field_name}}", {
                            #       "value": "",
                            #       "type": "",
                            #       "description": ""
                            #     }
                            column_name = "|".join(
                                [
                                    commons_url,
                                    indentifer_info.get("type"),
                                    identifier_name,
                                ]
                            )
                            crosswalk_columns.add(column_name)

                            crosswalk_info[
                                commons_url + "|" + identifier_name
                            ] = indentifer_info.get("description")

                    json.dump(guid_crosswalk_metadata, cached_guid_file)

        crosswalk_columns = sorted(list(crosswalk_columns))

        logging.debug(f"got columns: {crosswalk_columns}")
        logging.debug(f"got crosswalk_info: {crosswalk_info}")

        output_info_filename = "".join(output_filename.split(".")[:-1]) + "_info.csv"
        logging.debug(f"writing crosswalk to: {output_filename}...")
        with open(
            output_filename,
            "w+",
        ) as output_file:
            writer = csv.DictWriter(
                output_file,
                **{**BASE_CSV_PARSER_SETTINGS, "fieldnames": crosswalk_columns},
            )
            writer.writeheader()

            # read from the temporary cached crosswalk metadata to get the values
            for guid in sorted(os.listdir(metadata_cache_dir)):
                with open(f"{metadata_cache_dir}/{guid}") as f:
                    fetched_metadata = json.load(f)

                    output_metadata = {}
                    for column in crosswalk_columns:
                        (
                            commons_url,
                            identifier_type,
                            identifier_name,
                        ) = _get_crosswalk_columns_parts(column)
                        output_metadata[column] = (
                            fetched_metadata.get(GUID_TYPE)
                            .get(commons_url, {})
                            .get(identifier_name, {})
                            .get("value", "")
                        )

                    # output_metadata looks something like this:
                    # {
                    #   "commons_url_a|identifier_type_a|identifier_name_a": "A01-00888",
                    #   "commons_url_b|identifier_type_b|identifier_name_b": "phs002363.v1_RC-1358",
                    # }
                    writer.writerow(output_metadata)

        logging.info(f"done writing crosswalk to: {output_filename}")
        logging.debug(f"writing crosswalk info to: {output_info_filename}...")

        with open(
            output_info_filename,
            "w+",
        ) as output_info_file:
            # Output format of additional info in csv (including descriptions):
            #     {{commons_url}}, {{identifier_name}}, {{description}}
            info_columns = ["commons_url", "identifier_name", "description"]
            writer = csv.DictWriter(
                output_info_file,
                **{**BASE_CSV_PARSER_SETTINGS, "fieldnames": info_columns},
            )
            writer.writeheader()

            for commons_url_and_identifier_name, description in crosswalk_info.items():
                output_metadata = {}
                commons_url, identifier_name = commons_url_and_identifier_name.split(
                    "|"
                )
                output_metadata["commons_url"] = commons_url
                output_metadata["identifier_name"] = identifier_name
                output_metadata["description"] = description or ""

                writer.writerow(output_metadata)

        logging.info(f"done writing crosswalk info to: {output_info_filename}")


async def publish_crosswalk_metadata(
    auth,
    file,
    info_file=None,
    guid_type=GUID_TYPE,
    mapping_methodologies=None,
):
    """
    Publish crosswalk metadata from a tsv file
    """
    mapping_methodologies = mapping_methodologies or []

    mds = Gen3Metadata(auth_provider=auth)
    index = Gen3Index(auth_provider=auth)

    delimiter = "," if file.endswith(".csv") else "\t"

    # dictionary of { "commons url | identifier name": description }
    crosswalk_info = {}
    if info_file:
        with open(info_file, "rt", encoding="utf-8-sig") as csvfile:
            metadata_info_reader = csv.DictReader(csvfile, delimiter=delimiter)

            if (
                metadata_info_reader.fieldnames
                and len(metadata_info_reader.fieldnames) != 3
            ) or [
                "commons_url",
                "identifier_name",
                "description",
            ] != metadata_info_reader.fieldnames:
                error_msg = (
                    "Invalid Crosswalk Metadata Info file format. Please ensure 3 "
                    f"columns for info with headers: commons_url,identifier_name,description). "
                    f"Found: {metadata_info_reader.fieldnames}"
                )
                logging.error(error_msg)
                raise Exception(error_msg)

            for metadata_line in metadata_info_reader:
                # unique key of 2 first columns, value is the description (last column)
                crosswalk_info[
                    metadata_line["commons_url"].strip()
                    + "|"
                    + metadata_line["identifier_name"].strip()
                ] = metadata_line["description"].strip()

        logging.debug(f"crosswalk_info: {crosswalk_info}")

    with open(file, "rt", encoding="utf-8-sig") as csvfile:
        metadata_reader = csv.DictReader(csvfile, delimiter=delimiter)

        if len(metadata_reader.fieldnames) < 2:
            error_msg = (
                "Invalid Crosswalk Metadata file format. Please ensure > 2 "
                f"columns for mapping. Found: {len(metadata_reader.fieldnames)}"
            )
            logging.error(error_msg)
            raise Exception(error_msg)

        try:
            crosswalk_columns_parts = {
                columns.strip(): _get_crosswalk_columns_parts(columns)
                for columns in metadata_reader.fieldnames
            }
        except Exception as exc:
            error_msg = (
                "Invalid Crosswalk Metadata file format. Please ensure all "
                "columns are pipe-delimited with "
                "'commons_url | identifier_type | identifier_name'. "
                f"Found: {metadata_reader.fieldnames}"
            )
            logging.error(error_msg)
            raise

        logging.debug(f"crosswalk_columns_parts: {crosswalk_columns_parts}")

        pending_requests = []

        logging.debug(f"Attempting to get valid GUIDs...")
        guids_available_for_use = index.get_valid_guids(count=1000)
        logging.debug(f"Got {len(guids_available_for_use)} valid GUIDs for use.")

        for metadata_line in metadata_reader:
            raw_crosswalk_metadata = {
                key.strip(): value.strip() for key, value in metadata_line.items()
            }

            logging.debug(
                f"raw_crosswalk_metadata: {raw_crosswalk_metadata}, "
                f"from line: {metadata_line}"
            )

            metadata = {}
            metadata_aliases = []
            for column, value in raw_crosswalk_metadata.items():
                commons_url, identifier_type, identifier_name = crosswalk_columns_parts[
                    column
                ]
                metadata_aliases.append(value)

                to_update = {
                    "value": value,
                    "type": identifier_type,
                }
                description = crosswalk_info.get(
                    commons_url + "|" + identifier_name, ""
                )
                # only override the potentially existing description if a new one is
                # provided by the new crosswalk
                if description:
                    to_update.update({"description": description})

                metadata.setdefault(commons_url, {}).setdefault(
                    identifier_name, {}
                ).update(to_update)

            logging.debug(f"new crosswalk metadata: {metadata}")

            # MDS does not support a deep merge, so we need to merge any existing crosswalk
            # data with this new data here before updating

            # try to find any existing GUID from the values provided
            mds_record = None
            for alias in metadata_aliases:
                try:
                    mds_record = await mds.async_get(alias)
                except Exception as exc:
                    pass

                if mds_record:
                    break

            if mds_record:
                logging.debug(f"existing metadata record: {mds_record}")
                mds_record_guid_type = mds_record.get("_guid_type")
                mds_record_crosswalk_for_guid = mds_record.get(
                    CROSSWALK_NAMESPACE, {}
                ).get(guid_type, {})
                mds_record_mapping_methodologies = set(
                    mds_record.get(CROSSWALK_NAMESPACE, {})
                    .get(GUID_TYPE, {})
                    .get("mapping_methodologies", set())
                )
                for item in mapping_methodologies:
                    mds_record_mapping_methodologies.add(item)
            else:
                logging.debug(f"no existing metadata record found.")
                mds_record_guid_type = GUID_TYPE
                mds_record_crosswalk_for_guid = {}
                mds_record_mapping_methodologies = set(mapping_methodologies)

            # combine with new metadata using a deep update (e.g. merging nested dicts)
            deep_dict_update(mds_record_crosswalk_for_guid, metadata)
            mds_record_crosswalk_for_guid.update(
                {
                    "mapping_methodologies": sorted(
                        list(mds_record_mapping_methodologies)
                    )
                }
            )

            # add standard structure around the crosswalk metadata
            final_metadata = {
                "_guid_type": mds_record_guid_type,
                CROSSWALK_NAMESPACE: {
                    guid_type: mds_record_crosswalk_for_guid,
                },
            }

            # refresh list if needed, then get a guid
            if not guids_available_for_use:
                guids_available_for_use = index.get_valid_guids(count=1000)
            guid = guids_available_for_use.pop(-1)

            logging.info(f"crosswalk metadata for {guid}: {final_metadata}")

            # call update with merge to ensure this doesn't wipe out any
            # non-crosswalk namespaced blocks of metadata
            if mds_record:
                pending_requests += [
                    mds.async_update(
                        guid, final_metadata, aliases=metadata_aliases, merge=True
                    )
                ]
            else:
                pending_requests += [
                    mds.async_create(guid, final_metadata, aliases=metadata_aliases)
                ]

            if len(pending_requests) == MAX_CONCURRENT_REQUESTS:
                await asyncio.gather(*pending_requests)
                pending_requests = []

        await asyncio.gather(*pending_requests)


def try_delete_crosswalk_guid(auth, guid):
    """
    Deletes all crosswalk metadata under [guid] if it exists

    # TODO needs tests and CLI command
    """
    mds = Gen3Metadata(auth_provider=auth)
    try:
        metadata = mds.get(guid)
        if metadata["_guid_type"] == "crosswalk_metadata":
            mds.delete(guid)
        else:
            logging.warning(f"{guid} is not crosswalk metadata. Skipping.")
    except requests.exceptions.HTTPError as e:
        logging.warning(e)


def _get_crosswalk_columns_parts(column):
    commons_url, identifier_type, identifier_name = column.strip().split("|")
    return (
        commons_url.strip(),
        identifier_type.strip(),
        identifier_name.strip(),
    )



import aiohttp
import backoff
import requests
import urllib.parse
from cdislogging import get_logger

import sys

import indexclient.client as client

from gen3.utils import DEFAULT_BACKOFF_SETTINGS, raise_for_status_and_print_error
from gen3.auth import Gen3Auth

logging = get_logger("__name__")


class Gen3Index:
    """

    A class for interacting with the Gen3 Index services.

    Args:
        endpoint (str): public endpoint for reading/querying indexd - only necessary if auth_provider not provided
        auth_provider (Gen3Auth): A Gen3Auth class instance or indexd basic creds tuple

    Examples:
        This generates the Gen3Index class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... index = Gen3Index(auth)

    """

    def __init__(self, endpoint=None, auth_provider=None, service_location="index"):
        # legacy interface required endpoint as 1st arg
        if endpoint and isinstance(endpoint, Gen3Auth):
            auth_provider = endpoint
            endpoint = None
        if auth_provider and isinstance(auth_provider, Gen3Auth):
            endpoint = auth_provider.endpoint
        endpoint = endpoint.strip("/")
        # if running locally, indexd is deployed by itself without a location relative
        # to the commons
        if "http://localhost" in endpoint:
            service_location = ""

        if not endpoint.endswith(service_location):
            endpoint += "/" + service_location

        self.endpoint = endpoint
        self.client = client.IndexClient(endpoint, auth=auth_provider)

    ### Get Requests
    def is_healthy(self):
        """

        Return if indexd is healthy or not

        """
        try:
            response = self.client._get("_status")
            response.raise_for_status()
        except Exception:
            return False
        return response.text == "Healthy"

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_version(self):
        """

        Return the version of indexd

        """
        response = self.client._get("_version")
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_stats(self):
        """

        Return basic info about the records in indexd

        """
        response = self.client._get("_stats")
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_all_records(self, limit=None, paginate=False):
        """

        Get a list of all records

        """
        all_records = []
        url = "index/"

        if limit:
            url += f"?limit={limit}"

        response = self.client._get(url)
        raise_for_status_and_print_error(response)

        records = response.json().get("records")
        all_records.extend(records)

        if paginate and records:
            previous_did = None
            start_did = records[-1].get("did")

            while start_did != previous_did:
                previous_did = start_did

                params = {"start": f"{start_did}"}
                url_parts = list(urllib.parse.urlparse(url))
                query = dict(urllib.parse.parse_qsl(url_parts[4]))
                query.update(params)

                url_parts[4] = urllib.parse.urlencode(query)

                url = urllib.parse.urlunparse(url_parts)
                response = self.client._get(url)
                raise_for_status_and_print_error(response)

                records = response.json().get("records")
                all_records.extend(records)

                if records:
                    start_did = response.json().get("records")[-1].get("did")

        return all_records

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_records_on_page(self, limit=None, page=None):
        """

        Get a list of all records given the page and page size limit

        """
        params = {}
        url = "index/"

        if limit is not None:
            params["limit"] = limit

        if page is not None:
            params["page"] = page

        query = urllib.parse.urlencode(params)

        response = self.client._get(url + "?" + query)
        raise_for_status_and_print_error(response)

        return response.json().get("records")

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_record(self, guid=None, _ssl=None):
        """
        Asynchronous function to request a record from indexd.

        Args:
            guid (str): record guid

        Returns:
            dict: indexd record
        """
        url = f"{self.client.url}/index/{guid}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url, ssl=_ssl) as response:
                raise_for_status_and_print_error(response)
                response = await response.json()

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_records_on_page(self, limit=None, page=None, _ssl=None):
        """
        Asynchronous function to request a page from indexd.

        Args:
            page (int/str): indexd page to request

        Returns:
            List[dict]: List of indexd records from the page
        """
        all_records = []
        params = {}

        if limit is not None:
            params["limit"] = limit

        if page is not None:
            params["page"] = page

        query = urllib.parse.urlencode(params)

        url = f"{self.client.url}/index" + "?" + query
        async with aiohttp.ClientSession() as session:
            async with session.get(url, ssl=_ssl) as response:
                response = await response.json()

        return response.get("records")

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_records_from_checksum(
        self, checksum, checksum_type="md5", _ssl=None
    ):
        """
        Asynchronous function to request records from indexd matching checksum.

        Args:
            checksum (str): indexd checksum to request
            checksum_type (str): type of checksum, defaults to md5

        Returns:
            List[dict]: List of indexd records
        """
        all_records = []
        params = {}

        params["hash"] = f"{checksum_type}:{checksum}"

        query = urllib.parse.urlencode(params)

        url = f"{self.client.url}/index" + "?" + query
        async with aiohttp.ClientSession() as session:
            async with session.get(url, ssl=_ssl) as response:
                response = await response.json()

        return response.get("records")

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get(self, guid, dist_resolution=True):
        """

        Get the metadata associated with the given id, alias, or
        distributed identifier

        Args:
            guid: string
                - record id
            dist_resolution: boolean
            - *optional* Specify if we want distributed dist_resolution or not

        """
        rec = self.client.global_get(guid, dist_resolution)

        if not rec:
            return rec

        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_urls(self, size=None, hashes=None, guids=None):
        """

        Get a list of urls that match query params

        Args:
            size: integer
                - object size
            hashes: string
                - hashes specified as algorithm:value
            guids: list
                - list of ids

        """
        if guids:
            guids = ",".join(guids)
        p = {"size": size, "hash": hashes, "ids": guids}
        urls = self.client._get("urls", params=p).json()
        return [url for _, url in urls.items()]

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_record(self, guid):
        """

        Get the metadata associated with a given id

        """
        rec = self.client.get(guid)

        if not rec:
            return rec

        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_record_doc(self, guid):
        """

        Get the metadata associated with a given id

        """
        return self.client.get(guid)

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_with_params(self, params=None):
        """

        Return a document object corresponding to the supplied parameters, such
        as ``{'hashes': {'md5': '...'}, 'size': '...', 'metadata': {'file_state': '...'}}``.

            - need to include all the hashes in the request
            - index client like signpost or indexd will need to handle the
              query param `'hash': 'hash_type:hash'`

        """
        rec = self.client.get_with_params(params)

        if not rec:
            return rec

        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_get_with_params(self, params, _ssl=None):
        """

        Return a document object corresponding to the supplied parameter

        - need to include all the hashes in the request
        - need to handle the query param `'hash': 'hash_type:hash'`

        Args:
            params (dict): params to search with
            _ssl (None, optional): whether or not to use ssl

        Returns:
            Document: json representation of an entry in indexd

        """
        query_params = urllib.parse.urlencode(params)
        url = f"{self.client.url}/index/?{query_params}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url, ssl=_ssl) as response:
                await response.raise_for_status()
                response = await response.json()

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_latest_version(self, guid, has_version=False):
        """

        Get the metadata of the latest index record version associated
        with the given id

        Args:
            guid: string
                - record id
            has_version: boolean
                - *optional* exclude entries without a version

        """
        rec = self.client.get_latest_version(guid, has_version)

        if not rec:
            return rec

        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_versions(self, guid):
        """

        Get the metadata of index record version associated with the
        given id

        Args:
            guid: string
                - record id

        """
        response = self.client._get(f"/index/{guid}/versions")
        raise_for_status_and_print_error(response)
        versions = response.json()

        return [r for _, r in versions.items()]

    ### Post Requests

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_record(
        self,
        hashes,
        size,
        did=None,
        urls=None,
        file_name=None,
        metadata=None,
        baseid=None,
        acl=None,
        urls_metadata=None,
        version=None,
        authz=None,
        description=None,
        content_created_date=None,
        content_updated_date=None,
    ):
        """

        Create a new record and add it to the index

        Args:
            hashes (dict): {hash type: hash value,}
                eg ``hashes={'md5': ab167e49d25b488939b1ede42752458b'}``
            size (int): file size metadata associated with a given uuid
            did (str): provide a UUID for the new indexd to be made
            urls (list): list of URLs where you can download the UUID
            acl (list): access control list
            authz (list): RBAC strings
            file_name (str): name of the file associated with a given UUID
            metadata (dict): additional key value metadata for this entry
            urls_metadata (dict): metadata attached to each url
            baseid (str): optional baseid to group with previous entries versions
            version (str): entry version string
            description (str): optional description of the object
            content_created_date (datetime): optional creation date and time of the content being indexed
            content_updated_date (datetime): optional update date and time of the content being indexed
        Returns:
            Document: json representation of an entry in indexd

        """
        if urls is None:
            urls = []
        json = {
            "urls": urls,
            "hashes": hashes,
            "size": size,
            "file_name": file_name,
            "metadata": metadata,
            "urls_metadata": urls_metadata,
            "baseid": baseid,
            "acl": acl,
            "authz": authz,
            "version": version,
            "description": description,
            "content_created_date": content_created_date,
            "content_updated_date": content_updated_date,
        }
        if did:
            json["did"] = did
        rec = self.client.create(**json)

        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_create_record(
        self,
        hashes,
        size,
        did=None,
        urls=None,
        file_name=None,
        metadata=None,
        baseid=None,
        acl=None,
        urls_metadata=None,
        version=None,
        authz=None,
        _ssl=None,
        description=None,
        content_created_date=None,
        content_updated_date=None,
    ):
        """
        Asynchronous function to create a record in indexd.

        Args:
            hashes (dict): {hash type: hash value,}
                eg ``hashes={'md5': ab167e49d25b488939b1ede42752458b'}``
            size (int): file size metadata associated with a given uuid
            did (str): provide a UUID for the new indexd to be made
            urls (list): list of URLs where you can download the UUID
            acl (list): access control list
            authz (str): RBAC string
            file_name (str): name of the file associated with a given UUID
            metadata (dict): additional key value metadata for this entry
            urls_metadata (dict): metadata attached to each url
            baseid (str): optional baseid to group with previous entries versions
            version (str): entry version string
            description (str): optional description of the object
            content_created_date (datetime): optional creation date and time of the content being indexed
            content_updated_date (datetime): optional update date and time of the content being indexed

        Returns:
            Document: json representation of an entry in indexd
        """
        async with aiohttp.ClientSession() as session:
            if urls is None:
                urls = []

            json = {
                "form": "object",
                "hashes": hashes,
                "size": size,
                "urls": urls or [],
            }
            if did:
                json["did"] = did
            if file_name:
                json["file_name"] = file_name
            if metadata:
                json["metadata"] = metadata
            if baseid:
                json["baseid"] = baseid
            if acl:
                json["acl"] = acl
            if urls_metadata:
                json["urls_metadata"] = urls_metadata
            if version:
                json["version"] = version
            if authz:
                json["authz"] = authz
            if description:
                json["description"] = description
            if content_created_date:
                json["content_created_date"] = content_created_date
            if content_updated_date:
                json["content_updated_date"] = content_updated_date
            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self.client.auth._get_auth_value()}

            async with session.post(
                f"{self.client.url}/index/",
                json=json,
                headers=headers,
                ssl=_ssl,
            ) as response:
                assert response.status == 200, await response.json()
                response = await response.json()

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_blank(self, uploader, file_name=None):
        """

        Create a blank record

        Args:
            json - json in the format:
            {
                'uploader': type(string)
                'file_name': type(string) (optional*)
            }

        """
        json = {"uploader": uploader, "file_name": file_name}
        response = self.client._post(
            "index/blank",
            headers={"content-type": "application/json"},
            auth=self.client.auth,
            data=client.json_dumps(json),
        )
        raise_for_status_and_print_error(response)
        rec = response.json()

        return self.get_record(rec["did"])

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_new_version(
        self,
        guid,
        hashes,
        size,
        did=None,
        urls=None,
        file_name=None,
        metadata=None,
        acl=None,
        urls_metadata=None,
        version=None,
        authz=None,
        description=None,
        content_created_date=None,
        content_updated_date=None,
    ):
        """

        Add new version for the document associated to the provided uuid

        Since data content is immutable, when you want to change the
        size or hash, a new index document with a new uuid needs to be
        created as its new version. That uuid is returned in the did
        field of the response. The old index document is not deleted.

        Args:
            guid: (string): record id
            hashes (dict): {hash type: hash value,}
                eg ``hashes={'md5': ab167e49d25b488939b1ede42752458b'}``
            size (int): file size metadata associated with a given uuid
            did (str): provide a UUID for the new indexd to be made
            urls (list): list of URLs where you can download the UUID
            file_name (str): name of the file associated with a given UUID
            metadata (dict): additional key value metadata for this entry
            acl (list): access control list
            urls_metadata (dict): metadata attached to each url
            version (str): entry version string
            authz (str): RBAC string
            description (str): optional description of the object
            content_created_date (datetime): optional creation date and time of the content being indexed
            content_updated_date (datetime): optional update date and time of the content being indexed

            body: json/dictionary format
            - Metadata object that needs to be added to the store.
              Providing size and at least one hash is necessary and
              sufficient. Note: it is a good idea to add a version
              number

        """
        if urls is None:
            urls = []
        json = {
            "urls": urls,
            "form": "object",
            "hashes": hashes,
            "size": size,
            "file_name": file_name,
            "metadata": metadata,
            "urls_metadata": urls_metadata,
            "acl": acl,
            "authz": authz,
            "version": version,
            "description": description,
            "content_created_date": content_created_date,
            "content_updated_date": content_updated_date,
        }
        if did:
            json["did"] = did
        response = self.client._post(
            "index",
            guid,
            headers={"content-type": "application/json"},
            data=client.json_dumps(json),
            auth=self.client.auth,
        )
        raise_for_status_and_print_error(response)
        rec = response.json()

        if rec and "did" in rec:
            return self.get_record(rec["did"])
        return None

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_records(self, dids):
        """

        Get a list of documents given a list of dids

        Args:
            dids: list
                 - a list of record ids

        Returns:
            list: json representing index records

        """
        try:
            response = self.client._post(
                "bulk/documents", json=dids, auth=self.client.auth
            )
        except requests.HTTPError as exception:
            if exception.response.status_code == 404:
                return None
            else:
                raise exception

        return response.json()

    ### Put Requests

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def update_blank(self, guid, rev, hashes, size, urls=None, authz=None):
        """

        Update only hashes and size for a blank index

        Args:
            guid (string): record id
            rev (string): data revision - simple consistency mechanism
            hashes (dict): {hash type: hash value,}
                eg ``hashes={'md5': ab167e49d25b488939b1ede42752458b'}``
            size (int): file size metadata associated with a given uuid

        """
        params = {"rev": rev}
        json = {"hashes": hashes, "size": size}
        if urls:
            json["urls"] = urls
        if authz:
            json["authz"] = authz

        response = self.client._put(
            "index/blank",
            guid,
            headers={"content-type": "application/json"},
            params=params,
            auth=self.client.auth,
            data=client.json_dumps(json),
        )
        raise_for_status_and_print_error(response)
        rec = response.json()

        return self.get_record(rec["did"])

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def update_record(
        self,
        guid,
        file_name=None,
        urls=None,
        version=None,
        metadata=None,
        acl=None,
        authz=None,
        urls_metadata=None,
        description=None,
        content_created_date=None,
        content_updated_date=None,
    ):
        """

        Update an existing entry in the index

        Args:
             guid: string
                 - record id
             body: json/dictionary format
                 - index record information that needs to be updated.
                 - can not update size or hash, use new version for that

        """
        updatable_attrs = {
            "file_name": file_name,
            "urls": urls,
            "version": version,
            "metadata": metadata,
            "acl": acl,
            "authz": authz,
            "urls_metadata": urls_metadata,
            "description": description,
            "content_created_date": content_created_date,
            "content_updated_date": content_updated_date,
        }
        rec = self.client.get(guid)
        if not rec:
            raise ValueError(
                f"No indexd record found for GUID '{guid}' at '{self.endpoint}'"
            )
        for k, v in updatable_attrs.items():
            if v is not None:
                exec(f"rec.{k} = v")
        rec.patch()
        return rec.to_json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_update_record(
        self,
        guid,
        file_name=None,
        urls=None,
        version=None,
        metadata=None,
        acl=None,
        authz=None,
        urls_metadata=None,
        _ssl=None,
        description=None,
        content_created_date=None,
        content_updated_date=None,
        **kwargs,
    ):
        """
        Asynchronous function to update a record in indexd.

        Args:
             guid: string
                 - record id
             body: json/dictionary format
                 - index record information that needs to be updated.
                 - can not update size or hash, use new version for that
        """
        async with aiohttp.ClientSession() as session:
            updatable_attrs = {
                "file_name": file_name,
                "urls": urls,
                "version": version,
                "metadata": metadata,
                "acl": acl,
                "authz": authz,
                "urls_metadata": urls_metadata,
                "description": description,
                "content_created_date": content_created_date,
                "content_updated_date": content_updated_date,
            }
            record = await self.async_get_record(guid)
            revision = record.get("rev")

            for key, value in updatable_attrs.items():
                if value is not None:
                    record[key] = value

            del record["created_date"]
            del record["rev"]
            del record["updated_date"]
            del record["version"]
            del record["uploader"]
            del record["form"]
            del record["urls_metadata"]
            del record["baseid"]
            del record["size"]
            del record["hashes"]
            del record["did"]

            logging.info(f"PUT-ing record: {record}")

            # aiohttp only allows basic auth with their built in auth, so we
            # need to manually add JWT auth header
            headers = {"Authorization": self.client.auth._get_auth_value()}

            async with session.put(
                f"{self.client.url}/index/{guid}?rev={revision}",
                json=record,
                headers=headers,
                ssl=_ssl,
            ) as response:
                assert response.status == 200, await response.json()
                response = await response.json()

        return response

    ### Delete Requests

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def delete_record(self, guid):
        """

        Delete an entry from the index

        Args:
            guid: string
                - record id

        Returns: Nothing

        """
        rec = self.client.get(guid)
        if rec:
            rec.delete()
        return rec

    ### Query Requests

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def query_urls(self, pattern):
        """

        Query all record URLs for given pattern

        Args:
            pattern (str): pattern to match against indexd urls

        Returns:
            List[records]: indexd records with urls matching pattern
        """
        response = self.client._get(f"/_query/urls/q?include={pattern}")
        raise_for_status_and_print_error(response)
        return response.json()

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    async def async_query_urls(self, pattern, _ssl=None):
        """
        Asynchronous function to query urls from indexd.

        Args:
            pattern (str): pattern to match against indexd urls

        Returns:
            List[records]: indexd records with urls matching pattern
        """
        url = f"{self.client.url}/_query/urls/q?include={pattern}"
        async with aiohttp.ClientSession() as session:
            logging.debug(f"request: {url}")
            async with session.get(url, ssl=_ssl) as response:
                raise_for_status_and_print_error(response)
                response = await response.json()

        return response

    ## Mint GUID Requests

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_valid_guids(self, count=None):
        """
        Get a list of valid GUIDs without indexing
        Args:
            count (int): number of GUIDs to request
        Returns:
            List[str]: list of valid indexd GUIDs
        """
        url = "/guid/mint"
        if count:
            url += f"?count={count}"

        response = self.client._get(url)
        response.raise_for_status()
        return response.json().get("guids", [])

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def get_guids_prefix(self):
        """
        Get the prefix for GUIDs if there is one
        Returns:
            str: prefix for this instance
        """
        response = self.client._get("/guid/prefix")
        response.raise_for_status()
        return response.json().get("prefix")


def _print_func_name(function):
    return "{}.{}".format(function.__module__, function.__name__)


def _print_kwargs(kwargs):
    return ", ".join("{}={}".format(k, repr(v)) for k, v in list(kwargs.items()))



from gen3.tools.indexing.download_manifest import async_download_object_manifest
from gen3.tools.indexing.verify_manifest import async_verify_object_manifest
from gen3.tools.indexing.validate_manifest_format import is_valid_manifest_format
from gen3.tools.indexing.index_manifest import index_object_manifest



import click
import os.path
import json
import requests
import sys
import gen3.auth as auth_tool

from cdislogging import get_logger

logging = get_logger("__name__")


def stderr(*str):
    logging.error(*str, sys.stderr)


@click.command()
@click.option("--request", "request", help="HTTP Method - GET, PUT, POST, DELETE")
@click.option(
    "--data", "data", help="json data to post - read from file if starts with @"
)
@click.argument("path")
@click.pass_context
def curl(ctx, path, request=None, data=None):
    """Curl the endpoint with a token - ex: gen3 curl /user/user"""
    auth_provider = ctx.obj["auth_factory"].get()
    output = auth_provider.curl(path, request, data)
    logging.info(output.text)
    if output.status_code < 200 or output.status_code > 299:
        stderr("err status code %i" % output.status_code)
        sys.exit(1)


@click.command()
@click.pass_context
def endpoint(ctx):
    """Get the endpoint associated with the active authenticator"""
    logging.info(ctx.obj["auth_factory"].get().endpoint)


@click.command()
@click.pass_context
def get_access_token(ctx):
    """Get an access token suitable to pass as an Authorization header bearer"""
    logging.info(ctx.obj["auth_factory"].get().get_access_token())


@click.command()
@click.argument("token_file")
def token_decode(token_file):
    """Decode the given token file - may be "-" to indicate stdin"""
    if token_file == "-":
        tokenStr = sys.stdin.read()
    else:
        with open(token_file) as f:
            tokenStr = f.read()
    token = auth_tool.decode_token(tokenStr)
    logging.info(json.dumps(token, indent=2))


@click.command()
def wts_endpoint():
    """Get the wts endpoint"""
    logging.info(auth_tool.get_wts_endpoint())


@click.command()
def wts_list():
    """list the idp's available from the wts in a Gen3 workspace environment"""
    logging.info(json.dumps(auth_tool.get_wts_idps(), indent=2))


@click.group()
def auth():
    """Commands for authentication and authorization"""
    pass


auth.add_command(wts_endpoint, name="wts-endpoint")
auth.add_command(wts_list, name="wts-list")
auth.add_command(token_decode, name="token-decode")
auth.add_command(get_access_token, name="access-token")
auth.add_command(endpoint, name="endpoint")
auth.add_command(curl, name="curl")



import backoff
import requests
import urllib.parse
from cdislogging import get_logger

import sys
from urllib.parse import urlparse

from gen3.utils import (
    append_query_params,
    DEFAULT_BACKOFF_SETTINGS,
    raise_for_status_and_print_error,
)

logging = get_logger("__name__")


def wsurl_to_tokens(ws_urlstr):
    """Tokenize ws:/// paths - so ws:///@user/bla/foo returns ("@user", "bla/foo")"""
    urlparts = urlparse(ws_urlstr)
    if urlparts.scheme != "ws":
        raise Exception("invalid path {}".format(ws_urlstr))
    pathparts = [part for part in urlparts.path.split("/") if part]
    if len(pathparts) < 1:
        raise Exception("invalid path {}".format(ws_urlstr))
    return (pathparts[0], "/".join(pathparts[1:]))


@backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
def get_url(urlstr, dest_path):
    """Simple url fetch to dest_path with backoff"""
    res = requests.get(urlstr)
    raise_for_status_and_print_error(res)
    if dest_path == "-":
        sys.stdout.write(res.text)
    else:
        with open(dest_path, "wb") as f:
            f.write(res.content)


@backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
def put_url(urlstr, src_path):
    """Simple put src_path to url with backoff"""
    with open(src_path, "rb") as f:
        res = requests.put(urlstr, data=f)
    raise_for_status_and_print_error(res)


class Gen3WsStorage:
    """A class for interacting with the Gen3 workspace storage service.

    Examples:
        This generates the Gen3WsStorage class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(endpoint, refresh_file="credentials.json")
        ... wss = Gen3WsStorage(auth)
    """

    def __init__(self, auth_provider=None):
        """
        Initialization for instance of the class to setup basic endpoint info.

        Args:
            auth_provider (Gen3Auth, optional): Gen3Auth class to handle passing your
                token, required for admin endpoints
        """
        self._auth_provider = auth_provider

    @backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
    def upload_url(self, ws, wskey):
        """
        Get a upload url for the given workspace key

        Args:
          ws (string): name of the workspace
          wskey (string): key of the object in the workspace
        """
        wskey = wskey.lstrip("/")
        res = self._auth_provider.curl("/ws-storage/upload/{}/{}".format(ws, wskey))
        raise_for_status_and_print_error(res)
        return res.json()

    @backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
    def upload(self, src_path, dest_ws, dest_wskey):
        """
        Upload a local file to the specified workspace path
        """
        url = self.upload_url(dest_ws, dest_wskey)["Data"]
        put_url(url, src_path)

    @backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
    def download_url(self, ws, wskey):
        """
        Get a download url for the given workspace key

        Args:
          ws (string): name of the workspace
          wskey (string): key of the object in the workspace
        """
        wskey = wskey.lstrip("/")
        res = self._auth_provider.curl("/ws-storage/download/{}/{}".format(ws, wskey))
        raise_for_status_and_print_error(res)
        return res.json()

    def download(self, src_ws, src_wskey, dest_path):
        """
        Download a file from the workspace to local disk

        Args:
          src_ws (string): name of the workspace
          src_wskey (string): key of the object in the workspace
          dest_path (string): to download the file to
        """
        durl = self.download_url(src_ws, src_wskey)["Data"]
        get_url(durl, dest_path)

    def copy(self, src_urlstr, dest_urlstr):
        """
        Parse src_urlstr and dest_urlstr, and call upload or download
        as appropriate
        """
        if src_urlstr[0:3] == "ws:":
            if dest_urlstr[0:3] == "ws:":
                raise Exception(
                    "source and destination may not both reference a workspace"
                )
            pathparts = wsurl_to_tokens(src_urlstr)
            return self.download(pathparts[0], pathparts[1], dest_urlstr)
        if dest_urlstr[0:3] == "ws:":
            pathparts = wsurl_to_tokens(dest_urlstr)
            return self.upload(src_urlstr, pathparts[0], pathparts[1])
        raise Exception("source and destination may not both be local")

    @backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
    def ls(self, ws, wskey):
        """
        List the contents under the given workspace path

        Args:
          ws (string): name of the workspace
          wskey (string): key of the object in the workspace
        """
        wskey = wskey.lstrip("/")
        res = self._auth_provider.curl("/ws-storage/list/{}/{}".format(ws, wskey))
        raise_for_status_and_print_error(res)
        return res.json()

    def ls_path(self, ws_urlstr):
        """
        Same as ls - but parses ws_urlstr argument of form:
          ws:///workspace/key

        Args:
          ws (string): name of the workspace
          wskey (string): key of the object in the workspace
        """
        pathparts = wsurl_to_tokens(ws_urlstr)
        return self.ls(pathparts[0], pathparts[1])

    @backoff.on_exception(backoff.expo, requests.HTTPError, **DEFAULT_BACKOFF_SETTINGS)
    def rm(self, ws, wskey):
        """
        Remove the given workspace key

        Args:
          ws (string): name of the workspace
          wskey (string): key of the object in the workspace
        """
        wskey = wskey.lstrip("/")
        res = self._auth_provider.curl(
            "/ws-storage/list/{}/{}".format(ws, wskey), request="DELETE"
        )
        raise_for_status_and_print_error(res)
        return res.json()

    def rm_path(self, ws_urlstr):
        """
        Same as rm - but parses the ws_urlstr argument
        """
        pathparts = wsurl_to_tokens(ws_urlstr)
        return self.rm(pathparts[0], pathparts[1])



"""
Tools for ingesting a CSV/TSV metadata manifest into the Metdata Service.

Attributes:
    COLUMN_TO_USE_AS_GUID (str): file column containing guid to use
    GUID_TYPE_FOR_INDEXED_FILE_OBJECT (str): type to populate in mds when guid exists
        in indexd
    GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT (str): type to populate in mds when guid does
        NOT exist in indexd
    manifest_row_parsers (Dict{str: function}): functions for parsing, users can override
        manifest_row_parsers = {
            "guid_from_file": _get_guid_for_row,
            "indexed_file_object_guid": _query_for_associated_indexd_record_guid,
        }

        "guid_for_row" is the function to retrieve the guid from the given file
        "indexed_file_object_guid" is the function to retrieve the guid from elsewhere,
            like indexd (by querying)

    MAX_CONCURRENT_REQUESTS (int): Maximum concurrent requests to mds for ingestion
"""
import aiohttp
import asyncio
import csv
import json
from cdislogging import get_logger

import os
import time
import urllib.parse

from gen3.index import Gen3Index
from gen3.metadata import Gen3Metadata

TMP_FOLDER = os.path.abspath("./tmp") + "/"
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

MAX_CONCURRENT_REQUESTS = 24
COLUMN_TO_USE_AS_GUID = "guid"
GUID_TYPE_FOR_INDEXED_FILE_OBJECT = "indexed_file_object"
GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT = "metadata_object"

logging = get_logger("__name__")


def _get_guid_for_row(commons_url, row, lock):
    """
    Given a row from the manifest, return the guid to use for the metadata object.

    Args:
        commons_url (str): root domain for commons where mds lives
        row (dict): column_name:row_value
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections

    Returns:
        str: guid
    """
    return row.get(COLUMN_TO_USE_AS_GUID)


async def _query_for_associated_indexd_record_guid(
    commons_url, row, lock, output_queue
):
    """
    Given a row from the manifest, return the guid for the related indexd record.

    By default attempts to use a column "submitted_sample_id" to pattern match
    URLs in indexd records to find a single match.

    For example:
        "NWD12345" would match a record with url: "s3://some-bucket/file_NWD12345.cram"

    WARNING: The query endpoint this uses in indexd is incredibly slow when there are
             lots of indexd records.

    Args:
        commons_url (str): root domain for commons where mds lives
        row (dict): column_name:row_value
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
        output_queue (asyncio.Queue): queue for logging output

    Returns:
        str: guid or None
    """
    mapping = {"urls": "submitted_sample_id"}
    # Alternate example:
    #
    # mapping = {
    #     "acl": "study_with_consent",
    #     "size": "file_size",
    # }

    if "urls" in mapping and len(mapping.items()) > 1:
        msg = (
            "You cannot pattern match 'urls' and exact match other fields for mapping "
            "from indexd record to metadata columns. You can match by URL pattern "
            "*OR* match exact record fields like size, hash, uploader, url, acl, authz."
            f"\nYou provided mapping: {mapping}"
        )
        logging.error(msg)
        await output_queue.put(msg)
        raise Exception(msg)

    # special query endpoint for matching url patterns, other fields
    # just use get with params
    if "urls" in mapping:
        pattern = row.get(mapping["urls"])
        logging.debug(f"trying to find matching record matching url pattern: {pattern}")
        records = await async_query_urls_from_indexd(pattern, commons_url, lock)
    else:
        params = {
            mapping_key: row.get(mapping_value)
            for mapping_key, mapping_value in mapping.items()
        }
        logging.debug(f"trying to find matching record matching params: {params}")
        records = await _get_with_params_from_indexd(params, commons_url, lock)

    logging.debug(f"matching record(s): {records}")

    if len(records) > 1:
        msg = (
            "Multiple records were found with the given search criteria, this is assumed "
            "to be unintentional so the metadata will NOT be linked to these records:\n"
            f"{records}"
        )
        logging.warning(msg)
        await output_queue.put(msg)
        records = []

    guid = None
    if len(records) == 1:
        guid = records[0].get("did")

    return guid


manifest_row_parsers = {
    "guid_for_row": _get_guid_for_row,
    "indexed_file_object_guid": _query_for_associated_indexd_record_guid,
}


async def async_ingest_metadata_manifest(
    commons_url,
    manifest_file,
    metadata_source,
    auth=None,
    max_concurrent_requests=MAX_CONCURRENT_REQUESTS,
    manifest_row_parsers=manifest_row_parsers,
    manifest_file_delimiter=None,
    output_filename=f"ingest-metadata-manifest-errors-{time.time()}.log",
    get_guid_from_file=True,
    metadata_type=None,
):
    """
    Ingest all metadata records into a manifest csv

    Args:
        commons_url (str): root domain for commons where mds lives
        manifest_file (str): the file to ingest against
        metadata_source (str): the name of the source of metadata (used to namespace
            in the metadata service) ex: dbgap
        auth (Gen3Auth): Gen3 auth or tuple with basic auth name and password
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
        manifest_row_parsers (Dict{indexd_field:func_to_parse_row}): Row parsers
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str): filename for output logs
        get_guid_from_file (bool): whether or not to get the guid for metadata from file
            NOTE: When this is True, will use the function in
                  manifest_row_parsers["guid_for_row"] to determine the GUID
                  (usually just a specific column in the file row like "guid")
        metadata_type (str): the type of metadata to be filled into the _guid_type field.
            If provided, will override the default logic per GUID: (GUID_TYPE_FOR_INDEXED_FILE_OBJECT
                if is_indexed_file_object
                else GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT)
    """
    # if delimiter not specified, try to get based on file ext
    if not manifest_file_delimiter:
        file_ext = os.path.splitext(manifest_file)
        if file_ext[-1].lower() == ".tsv":
            manifest_file_delimiter = "\t"
        else:
            # default, assume CSV
            manifest_file_delimiter = ","

    await _ingest_all_metadata_in_file(
        commons_url,
        manifest_file,
        metadata_source,
        auth,
        manifest_file_delimiter,
        max_concurrent_requests,
        output_filename.split("/")[-1],
        get_guid_from_file,
        metadata_type,
    )


async def _ingest_all_metadata_in_file(
    commons_url,
    manifest_file,
    metadata_source,
    auth,
    manifest_file_delimiter,
    max_concurrent_requests,
    output_filename,
    get_guid_from_file,
    metadata_type,
):
    """
    Ingest metadata from file into metadata service. This function
    creates semaphores to limit the number of concurrent http connections that
    get opened to send requests to mds.

    It then uses asyncio to start a number of coroutines. Steps:
        1) reads given metadata file (writes resulting rows to a queue)
        2) posts/puts to mds to write metadata for rows

    Args:
        commons_url (str): root domain for commons where mds lives
        manifest_file (str): the file to ingest against
        metadata_source (str): the name of the source of metadata (used to namespace
            in the metadata service) ex: dbgap
        auth (Gen3Auth): Gen3 auth or tuple with basic auth name and password
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str): filename for output logs
        get_guid_from_file (bool): whether or not to get the guid for metadata from file
            NOTE: When this is True, will use the function in
                  manifest_row_parsers["guid_for_row"] to determine the GUID
                  (usually just a specific column in the file row like "guid")
        metadata_type (str): the type of metadata to be filled into the _guid_type field.
            If provided, will override the default logic per GUID: (GUID_TYPE_FOR_INDEXED_FILE_OBJECT
                if is_indexed_file_object
                else GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT)
    """
    max_requests = int(max_concurrent_requests)
    logging.debug(f"max concurrent requests: {max_requests}")
    lock = asyncio.Semaphore(max_requests)
    queue = asyncio.Queue()
    output_queue = asyncio.Queue()

    start_time = time.perf_counter()
    msg = f"start time: {start_time}"
    logging.info(msg)
    await output_queue.put(msg)

    with open(manifest_file, encoding="utf-8-sig") as manifest:
        reader = csv.DictReader(manifest, delimiter=manifest_file_delimiter)
        for row in reader:
            new_row = {}
            for key, value in row.items():
                # I know this looks crazy, DictReader is doing goofy things when
                # column contains a JSON-like string so we're trying to fix it here
                # Basically make sure the resulting column is something that we can
                # later json.loads().
                # remove redudant quoting
                if value:
                    value = value.strip().strip("'").strip('"').replace("''", "'")
                new_row[key.strip()] = value
            await queue.put(new_row)

    await asyncio.gather(
        *(
            _parse_from_queue(
                queue,
                lock,
                commons_url,
                output_queue,
                auth,
                get_guid_from_file,
                metadata_source,
                metadata_type,
            )
            # why "+ (max_concurrent_requests / 4)"?
            # This is because the max requests at any given time could be
            # waiting for metadata responses all at once and there's processing done
            # before that semaphore, so this just adds a few extra processes to get
            # through the queue up to that point of metadata requests so it's ready
            # right away when a lock is released. Not entirely necessary but speeds
            # things up a tiny bit to always ensure something is waiting for that lock
            for x in range(
                0, int(max_concurrent_requests + (max_concurrent_requests / 4))
            )
        )
    )

    end_time = time.perf_counter()
    msg = f"end time: {end_time}"
    logging.info(msg)
    await output_queue.put(msg)

    msg = f"run time: {end_time-start_time}"
    logging.info(msg)
    await output_queue.put(msg)

    output_filename = os.path.abspath(output_filename)
    logging.info(
        f"done processing, writing output queue to single file {output_filename}"
    )

    # remove existing output if it exists
    if os.path.isfile(output_filename):
        os.unlink(output_filename)

    with open(output_filename, "w") as outfile:
        while not output_queue.empty():
            line = await output_queue.get()
            outfile.write(line + "\n")

    logging.info(f"done writing output to file {output_filename}")


async def _parse_from_queue(
    queue,
    lock,
    commons_url,
    output_queue,
    auth,
    get_guid_from_file,
    metadata_source,
    metadata_type,
):
    """
    Keep getting items from the queue and checking if indexd contains a record with
    that guid. Then create/update metadta for that GUID in the metadata service.
    Also log to output queue. Return when nothing is left in the queue.

    Args:
        queue (asyncio.Queue): queue to read metadata from
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
        commons_url (str): root domain for commons where mds lives
        output_queue (asyncio.Queue): queue for logging output
        auth (Gen3Auth): Gen3 auth or tuple with basic auth name and password
        get_guid_from_file (bool): whether or not to get the guid for metadata from file
            NOTE: When this is True, will use the function in
                  manifest_row_parsers["guid_for_row"] to determine the GUID
                  (usually just a specific column in the file row like "guid")
        metadata_source (str): the name of the source of metadata (used to namespace
            in the metadata service) ex: dbgap
        metadata_type (str): the type of metadata to be filled into the _guid_type field.
            If provided, will override the default logic per GUID: (GUID_TYPE_FOR_INDEXED_FILE_OBJECT
                if is_indexed_file_object
                else GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT)
    """
    while not queue.empty():
        row = await queue.get()

        if get_guid_from_file:
            guid = manifest_row_parsers["guid_for_row"](commons_url, row, lock)
            is_indexed_file_object = await _is_indexed_file_object(
                guid, commons_url, lock
            )
        else:
            guid = await manifest_row_parsers["indexed_file_object_guid"](
                commons_url, row, lock, output_queue
            )
            is_indexed_file_object = True

        if guid:
            # construct metadata from rows, don't include redundant guid column
            logging.debug(f"row: {row}")
            metadata_from_file = {}

            for key, value in row.items():
                try:
                    new_value = json.loads(value)
                except json.decoder.JSONDecodeError as exc:
                    if "}" in value or "{" in value or "[" in value or "]" in value:
                        msg = (
                            f"Unable to json.loads a string that looks like json: {value}. "
                            f"adding as a string instead of nested json. Exception: {exc}"
                        )
                        logging.warning(msg)
                        await output_queue.put(msg)
                    new_value = value

                metadata_from_file[key] = new_value

            if COLUMN_TO_USE_AS_GUID in metadata_from_file.keys():
                del metadata_from_file[COLUMN_TO_USE_AS_GUID]

            logging.debug(f"metadata from file: {metadata_from_file}")

            # namespace by metadata source
            metadata = {metadata_source: metadata_from_file}

            if metadata_type:
                metadata["_guid_type"] = metadata_type
            else:
                metadata["_guid_type"] = (
                    GUID_TYPE_FOR_INDEXED_FILE_OBJECT
                    if is_indexed_file_object
                    else GUID_TYPE_FOR_NON_INDEXED_FILE_OBJECT
                )

            logging.debug(f"metadata: {metadata}")

            try:
                await _create_metadata(guid, metadata, auth, commons_url, lock)
                msg = f"Successfully created {guid}"
                logging.info(msg)
                await output_queue.put(msg)
            except Exception as exc:
                logging.debug(
                    f"Got conflict for {guid}. Let's update instead of create..."
                )
                await _update_metadata(guid, metadata, auth, commons_url, lock)
                msg = f"Successfully updated {guid}"
                logging.info(msg)
                await output_queue.put(msg)
        else:
            msg = (
                f"Did not add a metadata object for row because an invalid "
                f"GUID was parsed or no record with this GUID was found in "
                f"indexd: {guid}.\nRow: {row}"
            )
            logging.warning(msg)
            await output_queue.put(msg)


async def _create_metadata(guid, metadata, auth, commons_url, lock):
    """
    Gets a semaphore then creates metadata for guid

    Args:
        guid (str): indexd record globally unique id
        metadata (str): the metadata to add
        auth (Gen3Auth): Gen3 auth or tuple with basic auth name and password
        commons_url (str): root domain for commons where metadata service lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    mds = Gen3Metadata(commons_url, auth_provider=auth)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        response = await mds.async_create(guid, metadata, _ssl=ssl)
        return response


async def _update_metadata(guid, metadata, auth, commons_url, lock):
    """
    Gets a semaphore then updates metadata for guid

    Args:
        guid (str): indexd record globally unique id
        metadata (str): the metadata to add
        auth (Gen3Auth): Gen3 auth or tuple with basic auth name and password
        commons_url (str): root domain for commons where metadata service lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    mds = Gen3Metadata(commons_url, auth_provider=auth)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        response = await mds.async_update(guid, metadata, _ssl=ssl)
        return response


async def _is_indexed_file_object(guid, commons_url, lock):
    """
    Gets a semaphore then requests a record for the given guid

    Args:
        guid (str): indexd record globally unique id
        commons_url (str): root domain for commons where mds lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    index = Gen3Index(commons_url)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        try:
            record = await index.async_get_record(guid, _ssl=ssl)
        except Exception as exc:
            # if error, assume it does not exist
            return False

        return bool(record)


async def async_query_urls_from_indexd(pattern, commons_url, lock):
    """
    Gets a semaphore then requests a record for the given pattern

    Args:
        pattern (str): url pattern to match
        commons_url (str): root domain for commons where mds lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    index = Gen3Index(commons_url)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        return await index.async_query_urls(pattern, _ssl=ssl)


async def _get_with_params_from_indexd(params, commons_url, lock):
    """
    Gets a semaphore then requests a record for the given params

    Args:
        params (str): params to match
        commons_url (str): root domain for commons where mds lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    index = Gen3Index(commons_url)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        return await index.async_get_with_params(params, _ssl=ssl)



"""
Takes the differential between the matching columns of two manifests

TODO: Able to handle situations where only specific columns should be compared
"""

import os
import csv
import sys
import traceback

from cdislogging import get_logger

logging = get_logger("__name__")


def manifest_diff(
    directory=".",
    files=None,
    key_column="id",
    allow_additional_columns=False,
    output_manifest_file_delimiter=None,
    output_manifest="diff-manifest.tsv",
    **kwargs,
):
    """
    Args:
        directory(str):
            Path of the directory containing the input manifests. All of the manifests contained in directory are assumed to be in a delimiter-separated values (DSV) format, and that there are no other non-DSV files in directory.
        files(list[str]):
            List of paths containing the input manifests. All of the manifests contained in directory are assumed to be in a delimiter-separated values (DSV) format, and that there are no other non-DSV files in directory
        key_column(str):
            column of unique identifier in manifest
        manifest_content(dict(dict)):
            Dict of manifest content being compared
        allow_additional_columns(boolean):
            Bool allowing manifests to have different columns
        output_manifest_file_delimiter(str):
            The delimiter used for writing the output manifest. If not provided, the delimiter will be determined
            based on the file extension of output_manifest
        output_manifest(str):
            The file to write the output manifest to

    Returns:
        None
    """

    files = files or []
    if not files:
        logging.info(f"Iterating over manifests in {directory} directory")
        for file in sorted(os.listdir(directory)):
            files.append(os.path.join(directory, file))

    files.sort()

    content = _precheck_manifests(
        allow_additional_columns=allow_additional_columns,
        files=files,
        key_column=key_column,
    )

    diff_content = _compare_manifest_columns(
        allow_additional_columns=allow_additional_columns,
        manifest_content=content,
    )

    _write_csv(
        output_manifest_file_delimiter=output_manifest_file_delimiter,
        output_manifest=output_manifest,
        diff_content=diff_content,
    )


def _precheck_manifests(
    allow_additional_columns,
    key_column,
    files=[],
    **kwargs,
):
    """
    Precheck of all manifests for:
    - two files given
    - files share same extension
    - if additional columns not allowed, check for all headers are matching between two files

    Args:
        allow_additional_columns(bool)
        files(list[str])
        key_column(str)

    Returns:
        if pass all checks, dict(list, list): CSV content and headers
        else, bool False

        {
            csvdict:{}
                "id1": [{"header1": "", "header2": "", ...}],
                "id2": [{}],
                ...
            },
            headers: ["header1", "header2", ...]
        }
    """

    logging.info(f"Prechecking manifest files: {files}")

    if not len(files) == 2:
        raise Exception("Must take difference of two files, check dir for hidden files")

    tsv_files = [file_name for file_name in files if ".tsv" in file_name.lower()]
    csv_files = [file_name for file_name in files if ".csv" in file_name.lower()]
    if len(tsv_files) == len(files):
        file_delimiter = "\t"
    elif len(csv_files) == len(files):
        file_delimiter = ","
    else:
        raise ValueError("Not all files have the same extension type")

    headers = []
    manifest_content = []
    for manifest in files:
        with open(manifest, "r", encoding="utf-8-sig") as csvfile:
            csv_reader = csv.DictReader(csvfile, delimiter=file_delimiter)

            field_names = csv_reader.fieldnames
            logging.debug(f"Field names from {manifest}: {field_names}")
            headers.append(field_names)
            if len(headers) == 2:
                if not allow_additional_columns:
                    if not headers[0] == headers[1]:
                        raise Exception(
                            f"Headers are not the same among manifests. {set(headers[1]) ^ set(headers[0])}"
                        )

            content = {}
            for row in csv_reader:
                for column in row:
                    if column == "acl":
                        row[column] = str(sorted(eval(row[column])))
                content[row[key_column]] = row

            manifest_content.append(content)

    return {
        "csvdict": manifest_content,
        "headers": [x for x in headers[0] if x in headers[1]],
    }


def _compare_manifest_columns(
    allow_additional_columns,
    manifest_content={},
    **kwargs,
):
    """
    Args:
        manifest_content(dict(list, list)): Dict of CSV list and headers set
        allow_additional_columns(bool)

    Returns:
        Dict containing dict of diff list and headers list
    """

    headers = manifest_content["headers"]
    # TODO ability to only have certain headers compared
    dict1 = manifest_content["csvdict"][0]
    dict2 = manifest_content["csvdict"][1]

    diff_content = [dict2[i] for i in dict2 if i not in dict1 or dict2[i] != dict1[i]]

    return {"headers": headers, "csvdict": diff_content}


def _write_csv(output_manifest_file_delimiter, output_manifest, diff_content={}):
    """
    write to csv file

    Args:
        output_manifest_file_delimiter(str): file name
        output_manifest: file name with extension
        diff_content(dict(list)): dict of file info

    Returns:
        None.
        Writes a manifest file of the diff between manifests
    """

    if output_manifest_file_delimiter is None:
        output_manifest_file_ext = os.path.splitext(output_manifest)
        if output_manifest_file_ext[-1].lower() == ".tsv":
            output_manifest_file_delimiter = "\t"
        else:
            output_manifest_file_delimiter = ","

    headers = diff_content["headers"]

    logging.info(f"Writing diff manifest to {output_manifest}")
    with open(output_manifest, "w") as outputfile:
        output_writer = csv.DictWriter(
            outputfile,
            delimiter=output_manifest_file_delimiter,
            fieldnames=headers,
        )
        output_writer.writeheader()

        for record in diff_content["csvdict"]:
            output_writer.writerow(record)

        logging.info(f"Finished writing merged manifest to {output_manifest}")



from gen3.tools.metadata.ingest_manifest import async_ingest_metadata_manifest
from gen3.tools.metadata.ingest_manifest import async_query_urls_from_indexd
from gen3.tools.metadata.verify_manifest import async_verify_metadata_manifest



"""
This is a lower-level DOI class for interacting with Datacite and actually
minting the DOIs.

For collecting DOI Metadata, other classes (outside of the ones in this module)
can interact with different APIs to gather the necessary metadata.
"""
import backoff
import requests
import os

from cdislogging import get_logger

from gen3.tools.metadata.discovery import get_discovery_metadata
from gen3.metadata import Gen3Metadata
from gen3.utils import (
    raise_for_status_and_print_error,
    is_status_code,
    DEFAULT_BACKOFF_SETTINGS,
)

logging = get_logger("__name__")


class DataCiteDOIValidationError(Exception):
    pass


class DataCite(object):
    """
    For interacting with DataCite's API.
    https://support.datacite.org/reference/introduction

    Example usage:

    ```
    import os
    from requests.auth import HTTPBasicAuth
    import sys

    from gen3.doi import DataCite, DigitalObjectIdentifer

    PREFIX = "10.12345"

    def main():
        datacite = DataCite(
            auth_provider=HTTPBasicAuth(
                os.environ["DATACITE_USERNAME"], os.environ["DATACITE_PASSWORD"]
            ), use_prod=True
        )

        doi = DigitalObjectIdentifer(
            prefix=PREFIX,
            # identifier=f"{PREFIX}/test1",
            creators=["test"],
            titles=["test"],
            publisher="test",
            publication_year=2023,
            doi_type_general="Other",
            doi_type="Study Record",
            url="https://example.com",
            version="0.1",
            descriptions=[{"description": "this is a test resource"}],
            foobar="test",
        )

        datacite.create_doi(doi)

        datacite.persist_doi_metadata_in_gen3(doi)

    if __name__ == "__main__":
        main()
    ```
    """

    DEFAULT_DOI_ACCESS_INFORMATION = (
        "The data identified by this DOI is available through Gen3. Please refer "
        "to other documentation for more details on access. Alternatively, use the contact "
        "information provided on this page to request such information."
    )
    DEFAULT_DOI_CONTACT_INFORMATION = (
        "Unavailable. Please refer to other links or "
        "documentation on this site to determine an effective contact."
    )
    PRODUCTION_URL = "https://api.datacite.org"
    TEST_URL = "https://api.test.datacite.org"

    def __init__(
        self,
        api=None,
        auth_provider=None,
        use_prod=False,
    ):
        """
        Initiatilize.

        Args:
            api (str, optional): Override API with provided URL. Will ignore
                `use_prod` if this is provided
            auth_provider (Gen3Auth, optional): A Gen3Auth class instance or compatible interface
            use_prod (bool, optional): Default False. Whether or not to use the
                production Datacite URL.
        """
        self._auth_provider = auth_provider
        if api:
            self.api = api
        elif use_prod:
            self.api = DataCite.PRODUCTION_URL
        else:
            self.api = DataCite.TEST_URL

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def create_doi(self, doi):
        """
        Create DOI provided via DataCite's API

        Args:
            doi (gen3.doi.DigitalObjectIdentifier): DOI to create
        """
        payload = doi.as_dict()
        headers = {
            "Content-type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        }

        endpoint = self.api.rstrip("/") + "/dois"
        logging.info(f"POST-ing to {endpoint}...")
        logging.debug(f"Data payload to {endpoint}: {payload}")
        response = requests.post(
            endpoint, json=payload, auth=self._auth_provider, headers=headers
        )
        logging.debug(f"Response: {response.text}")
        raise_for_status_and_print_error(response)

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def read_doi(self, identifier):
        """
        Get DOI provided by DataCite's API.

        Args:
            identifier (str): DOI identifier to read
        """
        headers = {
            "Content-type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        }

        endpoint = self.api.rstrip("/") + f"/dois/{identifier}"
        logging.info(f"GET-ing to {endpoint}...")
        response = requests.get(endpoint, auth=self._auth_provider, headers=headers)
        logging.debug(f"Response: {response.text}")

        # 404's should not raise exceptions, instead returning nothing
        if is_status_code(response, "404"):
            return None

        raise_for_status_and_print_error(response)

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def update_doi(self, doi):
        """
        Update DOI provided via DataCite's API

        Args:
            doi (gen3.doi.DigitalObjectIdentifier): DOI to update
        """
        payload = doi.as_dict()
        identifier = doi.identifier
        headers = {
            "Content-type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        }

        endpoint = self.api.rstrip("/") + f"/dois/{identifier}"
        logging.info(f"PUT-ing to {endpoint}...")
        logging.debug(f"Data payload to {endpoint}: {payload}")
        response = requests.put(
            endpoint, json=payload, auth=self._auth_provider, headers=headers
        )
        logging.debug(f"Response: {response.text}")
        raise_for_status_and_print_error(response)

        return response

    @backoff.on_exception(backoff.expo, Exception, **DEFAULT_BACKOFF_SETTINGS)
    def delete_doi(self, identifier):
        """
        Delete DOI provided via DataCite's API

        Args:
            identifier (str): DOI identifier to delete
        """
        headers = {
            "Content-type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        }

        endpoint = self.api.rstrip("/") + f"/dois/{identifier}"
        logging.info(f"DELETE-ing to {endpoint}...")
        response = requests.delete(endpoint, auth=self._auth_provider, headers=headers)
        logging.debug(f"Response: {response.text}")
        raise_for_status_and_print_error(response)

        return response

    def persist_doi_metadata_in_gen3(
        self,
        guid,
        doi,
        auth,
        additional_metadata=None,
        prefix="",
    ):
        """
        Persist the DOI Metadata in Gen3. The default behavior is
        to use the Gen3 Discovery feature set, where there exists an entry
        for the DOI already in the Discovery Metadata under the `guid` provided.

        The DOI Metadata will be added under the existing `guid`'s `gen3_discovery`
        metadata block so that it is available to the Gen3 Discovery Page.

        Args:
            guid (str): globally unique ID for the metadata record
            doi (DigitalObjectIdentifier): DOI object with required metadata
            auth (TYPE): Description
            additional_metadata (Dict, optional): Any additional metadata fields
                to add to the DOI metadata into the Gen3 Metadata API.
                This can include overrides to the default landing page info
                for the following fields:
                    is_available (bool): True/False
                    disclaimer (str): Text to display at the top of the page
                        intended use is for tombstone pages, where the data is
                        no longer available and we need to note that
                    citation (str): By default: `Creator (PublicationYear).
                        Title. Version. Publisher. ResourceType. Identifier`
                    access_information (str): Text on how to access this data
                    access_information_link (str): HTTP link to additional info
                        about access
                    version_information (str): Text on any version information
                    contact (str): Contact information required for landing page
            prefix (str): prefix for the metadata fields to prepend to all DOI
                fields from `doi` AND all fields in `additional_metadata`
        """
        metadata_service = Gen3Metadata(auth_provider=auth)
        identifier = doi.identifier

        metadata = doi.as_gen3_metadata(prefix=prefix)

        additional_metadata = additional_metadata or {}

        is_available = additional_metadata.get("is_available", True)
        is_available_text = "Yes" if is_available else "No"
        metadata[prefix + "is_available"] = is_available_text

        if additional_metadata.get("is_available"):
            del additional_metadata["is_available"]

        metadata[prefix + "disclaimer"] = additional_metadata.get("disclaimer", "")
        if additional_metadata.get("disclaimer"):
            del additional_metadata["disclaimer"]

        metadata[prefix + "citation"] = additional_metadata.get("citation")
        if not metadata[prefix + "citation"]:
            if not doi.optional_fields.get("version"):
                logging.warning(
                    f"DOI {doi.identifier} is missing `version` in doi.optional_fields. "
                    f"Defaulting to `1` for citation."
                )
            # `Creator (PublicationYear). Title. Version. Publisher. ResourceType. Identifier`
            metadata[prefix + "citation"] = (
                f"{'& '.join([creator.get('name') for creator in doi.creators])} "
                f"({doi.publication_year}). "
                f"{'& '.join([titles.get('title') for titles in doi.titles])} "
                f"{str(doi.optional_fields.get('version', '1'))}. "
                f"{doi.publisher}. {doi.doi_type_general}. {doi.identifier}"
            )
        if additional_metadata.get("citation"):
            del additional_metadata["citation"]

        metadata[prefix + "version_information"] = additional_metadata.get(
            "version_information"
        )
        if not metadata[prefix + "version_information"]:
            if not doi.optional_fields.get("version"):
                logging.warning(
                    f"DOI {doi.identifier} is missing `version` in doi.optional_fields. "
                    f"Defaulting to `1`."
                )
            metadata[
                prefix + "version_information"
            ] = f"This is version {doi.optional_fields.get('version', '1')} of this {doi.doi_type_general}."
        if additional_metadata.get("version_information"):
            del additional_metadata["version_information"]

        metadata[prefix + "access_information"] = additional_metadata.get(
            "access_information"
        )
        if not metadata[prefix + "access_information"]:
            logging.warning(
                "No `access_information` provided in "
                "`additional_metadata` in doi.persist_metadata_in_gen3() call. "
                f"Consider providing this. Default will be used: {DataCite.DEFAULT_DOI_ACCESS_INFORMATION}"
            )
            metadata[
                prefix + "access_information"
            ] = DataCite.DEFAULT_DOI_ACCESS_INFORMATION
        if additional_metadata.get("access_information"):
            del additional_metadata["access_information"]

        metadata[prefix + "contact"] = additional_metadata.get("contact")
        if not metadata[prefix + "contact"]:
            logging.warning(
                "No `contact` provided in "
                "`additional_metadata` in doi.persist_metadata_in_gen3() call. "
                f"Consider providing this. Default will be used: {DataCite.DEFAULT_DOI_CONTACT_INFORMATION}"
            )
            metadata[prefix + "contact"] = DataCite.DEFAULT_DOI_CONTACT_INFORMATION
        if additional_metadata.get("contact"):
            del additional_metadata["contact"]

        # ensure prefix gets added to any additional metadata
        additional_metadata_with_prefix = {
            prefix + key.lstrip(prefix): value
            for key, value in additional_metadata.items()
        }

        metadata.update(additional_metadata_with_prefix)
        logging.debug(f"Metadata for {guid}: {metadata}")

        try:
            existing_record = metadata_service.get(guid=guid)

            # combine new metadata with existing metadata b/c MDS only supports
            # shallow merge at the moment
            existing_record.setdefault("gen3_discovery", {}).update(metadata)
            metadata = existing_record

            metadata_service.update(guid=guid, metadata=metadata, aliases=[identifier])

            logging.info(f"Updated existing Gen3 metadata record for {guid}.")
            logging.debug(
                f"Updated existing Gen3 metadata record for {guid}: {existing_record}"
            )
        except requests.exceptions.HTTPError as exc:
            if is_status_code(exc, "404"):
                logging.info(
                    f"Existing Gen3 Discovery metadata record with GUID "
                    f"`{guid}` not found. Creating..."
                )
                metadata = get_discovery_metadata(metadata)
                logging.debug(
                    f"Creating missing GUID `{guid}` with metadata: {metadata}."
                )
                metadata_service.create(
                    guid=guid, metadata=metadata, aliases=[identifier]
                )
            else:
                # non 404s should actually escalate
                raise

        return metadata


class DigitalObjectIdentifier(object):
    """
    Object representation of a DOI (including it's metadata).
    Convert to a DataCite-compatible payload by using `as_dict()`

    Example return:
        {
          "data": {
            "id": "10.5438/0012",
            "type": "dois",
            "attributes": {
              "event": "publish",
              "doi": "10.5438/0012",
              "creators": [{
                "name": "DataCite Metadata Working Group"
              }],
              "titles": [{
                "title": "DataCite Metadata Schema Documentation for the Publication and Citation of Research Data v4.0"
              }],
              "publisher": "DataCite e.V.",
              "publicationYear": 2016,
              "types": {
                "resourceTypeGeneral": "Text"
              },
              "url": "https://schema.datacite.org/meta/kernel-4.4/index.html",
              "schemaVersion": "http://datacite.org/schema/kernel-4"
            }
          }
        }
    """

    PRODUCTION_DOI_RESOLVER = "https://doi.org"
    TEST_DOI_RESOLVER = "https://handle.test.datacite.org"

    # to update to a later version, the classes that handle representing various
    # objects will need to change. Also it's possible other parts of this implementation
    # will need to change (including general types, how they should be representined in JSON, etc)
    # As such, this will remain "hard coded"
    SCHEMA_VERSION = "http://datacite.org/schema/kernel-4.4"

    RESOURCE_GENERAL_TYPES = [
        "Audiovisual",
        "Book",
        "BookChapter",
        "Collection",
        "ComputationalNotebook",
        "ConferencePaper",
        "ConferenceProceeding",
        "DataPaper",
        "Dataset",
        "Dissertation",
        "Event",
        "Image",
        "InteractiveResource",
        "Journal",
        "JournalArticle",
        "Model",
        "OutputManagementPlan",
        "PeerReview",
        "PhysicalObject",
        "Preprint",
        "Report",
        "Service",
        "Software",
        "Sound",
        "Standard",
        "Text",
        "Workflow",
        "Other",
    ]

    OPTIONAL_FIELDS = [
        "subjects",
        "contributors",
        "descriptions",
        "dates",
        "relatedIdentifiers",
        "geoLocations",
        "language",
        "alternateIdentifiers",
        "identifiers",
        "sizes",
        "formats",
        "version",
        "rightsList",
        "fundingReferences",
    ]

    DEFAULT_DOI_TYPE = "Dataset"

    def __init__(
        self,
        prefix=None,
        identifier=None,
        creators=None,
        titles=None,
        publisher=None,
        publication_year=None,
        doi_type_general=None,
        doi_type=None,
        url=None,
        root_url=None,
        event=None,
        resolver=None,
        use_prod=False,
        _type="dois",
        **kwargs,
    ):
        """
        To create a DOI you must specify identifier OR prefix.

        Mandatory fields for registration are defaulted the init.

        Supply any additional reccomended/optional as keyword args

        Arg definitions summarized from the official schema for v4.4:
        https://schema.datacite.org/meta/kernel-4.4/doc/DataCite-MetadataKernel_v4.4.pdf

        Args:
            prefix (str, optional): DOI Prefix
            identifier (str, optional): a unique string that identifies the resource (the DOI)
                NOTE: `identifier` is NOT case sensitive but the `url` is. Different
                      Datacite API's tend to return the DOI either with the mixed cases
                      presented, fully lower, or fully upper (which adds to the confusion).
                      As such, anything provided/received will be forced UPPER_CASE
                      automatically to reduce confusion.
            creators (List[Dict], optional): the main
                researcher(s) involved in producing the data,
                or the author(s) of the publication.
            titles (List[Dict], optional): a name or titles by which a resource is known
                Format [{"title": "some title"}, ...]
            publisher (str, optional): The name of the entity that holds, archives, publishes
                prints, distributes, releases, issues, or produces the resource.
                This property will be used to formulate the citation, so consider the
                prominence of the role. For software, use Publisher for the code
                repository. If there is an entity other than a code repository,
                that "holds, archives, publishes, prints, distributes, releases,
                issues, or produces" the code, use the property
                Contributor/contributorType/ hostingInstitution for the code repository.
            publication_year (int, optional): the year the data or resource was or will be made
                publicly available
            doi_type_general (str, optional): a description of the general type of a resource. (controlled list values, see https://support.datacite.org/docs/datacite-metadata-schema-v44-mandatory-properties#10a-resourcetypegeneral)
            doi_type (str, optional): a description of the resource (free-format text)
            url (str, optional): the web address of the landing page for the resource
            root_url (str, optional): the root url for the landing page, must be used
                in combination with `identifier` to form a final landing page url
            event (str, optional): what event to issue against the DOI
                Event field logic per DataCite API:
                    publish - Triggers a state move from draft or registered to findable
                    register - Triggers a state move from draft to registered
                    hide - Triggers a state move from findable to registered
            resolver (str, optional): Override resolver API with provided URL.
                Will ignore `use_prod` if this is provided. Default to DataCite
                prod / test resolvers depending on `use_prod` var.
            use_prod (bool, optional): whether or not to use the production Datacite
                urls
            _type (str, optional): The type of thing to interact with in Datacite
        """
        self.prefix = prefix
        self.identifier = identifier or ""
        self.creators = creators or []
        self.titles = titles or []
        self.publisher = publisher
        self.publication_year = publication_year
        self.doi_type_general = (
            doi_type_general or DigitalObjectIdentifier.DEFAULT_DOI_TYPE
        )
        self.doi_type = doi_type
        self.url = url
        self.root_url = root_url
        self.event = event
        self._type = _type

        # See NOTE: in docstring under `identifier`
        self.identifier = self.identifier.upper()

        # any additional kwargs get interpreted as optional fields
        self.optional_fields = {}

        if resolver:
            self.resolver = resolver
        elif use_prod:
            self.resolver = DigitalObjectIdentifier.PRODUCTION_DOI_RESOLVER
        else:
            self.resolver = DigitalObjectIdentifier.TEST_DOI_RESOLVER

        for key, value in kwargs.items():
            if key in DigitalObjectIdentifier.OPTIONAL_FIELDS:
                self.optional_fields[key] = value
            else:
                logging.warning(
                    f"Skipping '{key}={value}' because '{key}' "
                    f"is not a supported optional DOI metadata field."
                )
                logging.debug(
                    f"Supported optional DOI metadata fields: "
                    f"{DigitalObjectIdentifier.OPTIONAL_FIELDS}"
                )

        if not (self.identifier or self.prefix):
            raise DataCiteDOIValidationError(
                "You must either specify a) `identifier` which is a full DOI with prefix "
                "or b) `prefix` (which will let Datacite autogenerate the rest of "
                "the ID. You can't leave both blank."
            )

        if self.root_url and self.url:
            logging.warning(
                "You provided both `root_url` and `url`. "
                f"Preferring `url` ({self.url}) and ignoring `root_url` ({self.root_url})."
            )
            self.root_url = None

        if self.root_url and self.prefix and not self.identifier:
            raise DataCiteDOIValidationError(
                "To use `root_url`, you MUST specify `identifier` (which is a "
                "full DOI with prefix). You CANNOT rely on `prefix` (which will "
                "let Datacite autogenerate the rest of the ID)."
            )

        if not self.url:
            if not self.root_url:
                raise DataCiteDOIValidationError(
                    "You must either specify a) `url` which is a full landing page url  "
                    "or b) `root_url` (which will automatically get appended with "
                    "the `identifier` to form final url). You can't leave both blank."
                )

            self.url = self._get_url_from_root()

        if (
            doi_type_general
            and doi_type_general not in DigitalObjectIdentifier.RESOURCE_GENERAL_TYPES
        ):
            logging.error(f"{doi_type_general} is not an accepted resourceTypeGeneral")
            raise DataCiteDOIValidationError(
                f"{doi_type_general} is not an accepted resourceTypeGeneral"
            )

    def _get_url_from_root(self):
        return self.root_url.rstrip("/") + f"/{self.identifier}/"

    def as_gen3_metadata(self, prefix=""):
        """
        Return a flat dictionary for ingestion into Gen3's Metadata Service

        Returns:
            Dict: flat dictionary for ingestion into Gen3's Metadata Service
        """
        data = {}

        if self.identifier:
            data[prefix + "identifier"] = self.identifier
            data[prefix + "resolvable_link"] = (
                self.resolver.rstrip("/") + "/" + self.identifier
            )
        if self.creators:
            data[prefix + "creators"] = " & ".join(
                [item["name"] for item in self.creators]
            )
        if self.titles:
            data[prefix + "titles"] = ", ".join([item["title"] for item in self.titles])
        if self.publisher:
            data[prefix + "publisher"] = self.publisher
        if self.publication_year:
            data[prefix + "publication_year"] = self.publication_year
        if self.doi_type_general:
            data[prefix + "resource_type_general"] = self.doi_type_general
        if self.doi_type:
            data[prefix + "resource_type"] = self.doi_type
        if self.url:
            data[prefix + "url"] = self.url

        for key, value in self.optional_fields.items():
            if key == "contributors":
                data[prefix + key] = " & ".join([item["name"] for item in value])
            elif key == "descriptions":
                data[prefix + key] = "\n\n".join(
                    [
                        item["descriptionType"] + ": " + item["description"]
                        for item in value
                    ]
                )
            elif key == "alternateIdentifiers":
                # sometimes this is a list and sometimes is an object depending
                # on which Datacite API you hit (PUT/POST/DELETE). Handle both
                if type(value) == list:
                    value = ", ".join(
                        [
                            item.get("alternateIdentifier")
                            + f" ({item.get('alternateIdentifierType')})"
                            for item in value
                        ]
                    )
                else:
                    value = (
                        value.get("alternateIdentifier")
                        + f" ({value.get('alternateIdentifierType')})"
                    )

                data[prefix + key] = value
            elif key == "identifiers":
                # sometimes this is a list and sometimes is an object depending
                # on which Datacite API you hit (PUT/POST/DELETE). Handle both
                if type(value) == list:
                    value = ", ".join(
                        [
                            item.get("identifier") + f" ({item.get('identifierType')})"
                            for item in value
                        ]
                    )
                else:
                    value = (
                        value.get("identifier") + f" ({value.get('identifierType')})"
                    )

                data[prefix + key] = value
            elif key == "fundingReferences":
                # sometimes this is a list and sometimes is an object depending
                # on which Datacite API you hit (PUT/POST/DELETE). Handle both
                if type(value) == list:
                    value = ", ".join([funder.get("funderName") for funder in value])
                else:
                    value = value.get("funderName")

                data[prefix + key] = value
            else:
                data[prefix + key] = value

        return data

    def as_dict(self):
        data = {
            "data": {
                "id": self.identifier,
                "type": self._type,
                "attributes": {
                    "doi": self.identifier,
                },
            }
        }

        if self.creators:
            data["data"]["attributes"]["creators"] = [item for item in self.creators]
        if self.titles:
            data["data"]["attributes"]["titles"] = self.titles
        if self.publisher:
            data["data"]["attributes"]["publisher"] = self.publisher
        if self.publication_year:
            data["data"]["attributes"]["publicationYear"] = self.publication_year
        if self.doi_type_general:
            data["data"]["attributes"]["types"] = {
                "resourceTypeGeneral": self.doi_type_general
            }
        else:
            data["data"]["attributes"]["types"] = {}
        if self.doi_type:
            data["data"]["attributes"]["types"].update({"resourceType": self.doi_type})
        if self.url:
            data["data"]["attributes"]["url"] = self.url
        if self.event:
            data["data"]["attributes"]["event"] = self.event

        data["data"]["attributes"]["schemaVersion"] = self.SCHEMA_VERSION

        for key, value in self.optional_fields.items():
            data["data"]["attributes"][key] = value

        # per DataCite API's spec, if you don't specify a full DOI you need
        # to specify the prefix at least
        if not self.identifier:
            del data["data"]["attributes"]["doi"]
            del data["data"]["id"]
            data["data"]["attributes"]["prefix"] = self.prefix

        return data

    @staticmethod
    def from_datacite_create_doi_response(response, resolver=None, use_prod=False):
        """
        Return a DigitalObjectIdentifier instance from the Response from a
        Datacite Create DOI API call.

        Args:
            resolver (str, optional): override default URLs for the DOI resolution
                (which gets added to the metadata). If supplied, will override any default
                Datacite URLs regardless of `use_prod`
            use_prod (bool, optional): whether or not to use the production Datacite
                urls

        Returns:
            TYPE: Description
        """
        raw_attributes = response.json().get("data", {}).get("attributes", {})
        attributes = {}

        # we only want the mandatory and optional fields we know how to handle,
        # exclude all other fields returned by the API
        for key, value in raw_attributes.items():
            if key in DigitalObjectIdentifier.OPTIONAL_FIELDS:
                attributes[key] = value

        return DigitalObjectIdentifier(
            prefix=raw_attributes.get("prefix"),
            # Why .upper()? See NOTE: in docstring under `identifier`
            identifier=raw_attributes.get("doi").upper(),
            creators=raw_attributes.get("creators"),
            titles=raw_attributes.get("titles"),
            publisher=raw_attributes.get("publisher"),
            publication_year=raw_attributes.get("publicationYear"),
            url=raw_attributes.get("url"),
            doi_type=raw_attributes.get("types", {}).get("resourceType", None),
            doi_type_general=raw_attributes.get("types", {}).get(
                "resourceTypeGeneral", None
            ),
            resolver=resolver,
            use_prod=use_prod,
            **attributes,
        )


class DigitalObjectIdentifierCreator(object):
    """
    Class to help with representing a Creator object for the DataCite API
    per their schema.

    NOTE: If you use anything other than the schema this was built for (v4.4),
          this logic will need to be updated.

    From the DataCite Metadata Schema V 4.4:

    * 20.2 Creator 0-n
        The institution or person
        responsible for creating the
        related resource
        To supply multiple creators,
        repeat this property.

    * 20.2.1 creatorName 1
        The full name of the related
        item creator

        Examples: Charpy, Antoine;
        Jemison, Mae; Foo Data
        Center Note: The personal
        name, format should be:
        family, given. Non- roman
        names may be transliterated
        according to the ALA-LC
        schemas12.

    * 20.2.1.a nameType 0-1
        The type of name Controlled List Values:
        Organizational
        Personal (default)

    * 20.2.2 givenName 0-1
        The personal or first name of
        the creator
        Examples based on the 20.11.1 names: Antoine; Mae

    * 20.2.3 familyName 0-1
        The surname or last name of
        the creator

        Examples based on the 2.1
        names: Charpy; Jemison
    """

    NAME_TYPE_PERSON = "Personal"
    NAME_TYPE_ORGANIZATION = "Organizational"

    def __init__(
        self,
        name,
        name_type=NAME_TYPE_PERSON,
        given_name=None,
        family_name=None,
    ):
        self.name = name

        allowed_name_types = [
            DigitalObjectIdentifierCreator.NAME_TYPE_PERSON,
            DigitalObjectIdentifierCreator.NAME_TYPE_ORGANIZATION,
        ]
        if name_type not in allowed_name_types:
            logging.error(
                f"Provided name_type '{name_type}' is NOT in {allowed_name_types}. "
                f"Defaulting to '{DigitalObjectIdentifierCreator.NAME_TYPE_PERSON}'."
            )
            name_type = DigitalObjectIdentifierCreator.NAME_TYPE_PERSON

        self.name_type = name_type
        self.given_name = given_name
        self.family_name = family_name

    def as_dict(self):
        data = {
            "name": self.name,
            "nameType": self.name_type,
        }

        if self.given_name:
            data.update({"givenName": self.given_name})

        if self.family_name:
            data.update({"familyName": self.family_name})

        if (
            self.given_name
            and self.family_name
            and self.name_type == DigitalObjectIdentifierCreator.NAME_TYPE_PERSON
        ):
            data.update({"name": f"{self.family_name}, {self.given_name}"})

        return data


class DigitalObjectIdentifierFundingReference(object):
    """
    This could theoretically support more options, but at the point of writing,
    they weren't deemed necessary. Leaving this as an isolated class in case
    we want to expand in the future.

    From the DataCite Metadata Schema V 4.4:

    * 19 FundingReference 0-n
        Information about financial
        support (funding) for the
        resource being registered
        It is a best practice to supply
        funding information when
        financial support has been
        received.

    * 19.1 funderName 1
        Name of the funding provider Example: Gordon and Betty
        Moore Foundation

    * 19.2 funderIdentifier 0-1
        Uniquely identifies a funding
        entity, according to various
        types.
        Example:
        https://doi.org/10.13039/100
        000936

    * 19.2.a funderIdentifierType 0-1
        The type of the funderIdentifier Controlled List Values:
            GRID
            ISNI
            ROR
            Other
    """

    def __init__(
        self,
        name,
    ):
        self.name = name

    def as_dict(self):
        data = {
            "funderName": f"{self.name}",
        }

        return data


class DigitalObjectIdentifierTitle(object):
    """
    This could theoretically support more options, but at the point of writing,
    they weren't deemed necessary. Leaving this as an isolated class in case
    we want to expand in the future.

    From the DataCite Metadata Schema V 4.4:
    * 3 Title 1-n
        A name or title by which a
        resource is known. May be
        the title of a dataset or the
        name of a piece of software.
        Free text.

    * 3.a titleType 0-1
        The type of Title (other than
        the Main Title)
        Controlled List Values:
        AlternativeTitle
        Subtitle
        TranslatedTitle
        Other
    """

    def __init__(
        self,
        title,
    ):
        self.title = title

    def as_dict(self):
        data = {
            "title": f"{self.title}",
        }

        return data


class DigitalObjectIdentifierDescription(object):
    """
    From the DataCite Metadata Schema V 4.4:
    * 17 Description 0-n
        All additional information that
        does not fit in any of the other
        categories. May be used for
        technical information.
        Free text
        ***
        It is a best practice to supply a
        description.

    * 17.a descriptionType 1
        The type of the Description If Description is used,
        descriptionType is mandatory.

        Controlled List Values:
        Abstract
        Methods
        SeriesInformation
        TableOfContents
        TechnicalInfo
        Other

        Note: SeriesInformation as a
        container for series title,
        volume, issue, page number,
        and related fields, is now
        superseded by the new
        RelatedItem property
    """

    TYPE_ABSTRACT = "Abstract"
    TYPE_METHODS = "Methods"
    TYPE_SERIES_INFORMATION = "SeriesInformation"
    TYPE_TABLE_OF_CONTENTS = "TableOfContents"
    TYPE_TECHNICAL_INFO = "TechnicalInfo"
    TYPE_OTHER = "Other"

    def __init__(self, description, description_type=TYPE_ABSTRACT):
        self.description = description

        allowed_types = [
            DigitalObjectIdentifierDescription.TYPE_ABSTRACT,
            DigitalObjectIdentifierDescription.TYPE_METHODS,
            DigitalObjectIdentifierDescription.TYPE_SERIES_INFORMATION,
            DigitalObjectIdentifierDescription.TYPE_TABLE_OF_CONTENTS,
            DigitalObjectIdentifierDescription.TYPE_TECHNICAL_INFO,
            DigitalObjectIdentifierDescription.TYPE_OTHER,
        ]
        if description_type not in allowed_types:
            logging.error(
                f"Provided description_type '{description_type}' is NOT in {allowed_types}. "
                f"Defaulting to '{DigitalObjectIdentifierDescription.TYPE_ABSTRACT}'."
            )
            description_type = DigitalObjectIdentifierDescription.TYPE_ABSTRACT

        self.description_type = description_type

    def as_dict(self):
        data = {
            "description": f"{self.description}",
            "descriptionType": f"{self.description_type}",
        }

        return data


class DigitalObjectIdenfitierAlternateID(object):
    """
    From the DataCite Metadata Schema V 4.4:

    11 alternateIdentifier 0-n An identifier other than the
      primary Identifier applied to the
      resource being registered. This
      may be any alphanumeric string
      which is unique within its
      domain of issue. May be used for
      local identifiers. The
      AlternateIdentifier should be an
      additional identifier for the same
      instance of the resource (i.e.,
      same location, same file).
      Free text
      ***
      Example:
      E-GEOD-34814

    11.a alternateIdentifierType
      1 The type of the
      AlternateIdentifier
      Free text
      ***
      If alternateIdentifier is used,
      alternateIdentifierType is
      mandatory. For the above
      example, the
      alternateIdentifierType would
      be "A local accession number"
    """

    def __init__(self, alternate_id, alternate_id_type):
        self.alternate_id = alternate_id
        self.alternate_id_type = alternate_id_type

    def as_dict(self):
        data = {
            "alternateIdentifier": f"{self.alternate_id}",
            "alternateIdentifierType": f"{self.alternate_id_type}",
        }

        return data
